#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import abc


class IUid(abc.ABC):
    @abc.abstractmethod
    def uid(self):
        pass


class IUidLink(abc.ABC):
    @abc.abstractmethod
    def subject_uid(self):
        pass

    @abc.abstractmethod
    def object_uid(self):
        pass


class UidLink(IUidLink):
    def __init__(self, subject_uid, object_uid):
        self._subject_uid = subject_uid
        self._object_uid = object_uid

    def subject_uid(self):
        return self._subject_uid

    def object_uid(self):
        return self._object_uid


class IRecordStorage(abc.ABC):
    @abc.abstractmethod
    def load(self):
        pass

    @abc.abstractmethod
    def store(self, record):
        pass

    @abc.abstractmethod
    def delete(self):
        pass


class IEntityStorage(abc.ABC):
    @abc.abstractmethod
    def get_entity(self, uid):
        pass

    @abc.abstractmethod
    def get_all(self):
        pass

    @abc.abstractmethod
    def put_entities(self, entities):
        pass

    @abc.abstractmethod
    def delete_uids(self, uids):
        pass


class ILinkStorage(abc.ABC):
    @abc.abstractmethod
    def put_links(self, links):
        pass

    @abc.abstractmethod
    def delete_links(self, links):
        pass

    @abc.abstractmethod
    def delete_links_for_subjects(self, subject_uids):
        pass

    @abc.abstractmethod
    def delete_links_for_objects(self, object_uids):
        pass

    @abc.abstractmethod
    def get_links_for_subject(self, subject_uid):
        pass

    @abc.abstractmethod
    def get_links_for_object(self, object_uid):
        pass

    @abc.abstractmethod
    def get_all_links(self):
        pass



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import logging
import os
import re

from typing import Optional
from .. import api, crypto, utils
from ..params import KeeperParams

from .base import GroupCommand, dump_report_data
from .enterprise import EnterpriseCommand

from ..proto import automator_pb2 as automator_proto
from ..proto import ssocloud_pb2 as ssocloud

automator_list_parser = argparse.ArgumentParser(prog='automator-list')

automator_view_parser = argparse.ArgumentParser(prog='automator-view')
automator_view_parser.add_argument('target', help='Automator ID or Name.')

automator_create_parser = argparse.ArgumentParser(prog='automator-create')
automator_create_parser.add_argument('--name', dest='name', required=True, action='store', help='Automator name.')
automator_create_parser.add_argument('--node', dest='node', required=True, help='Node Name or ID.')


automator_edit_parser = argparse.ArgumentParser(prog='automator-edit')
automator_edit_parser.add_argument('target', help='Automator ID or Name.')
automator_edit_parser.add_argument('--name', dest='name', action='store', help='Automator name.')
automator_edit_parser.add_argument('--url', dest='url', action='store', help='Automator Webhook URL.')
automator_edit_parser.add_argument('--skill', dest='skill', action='append', choices=['device', 'team'],
                                   help='Automator Skills.')
automator_edit_parser.add_argument('--set', dest='setting', metavar="KEY=VALUE", action='append',
                                   help='Automator Settings. Use value: '
                                        '${file:filename} to load text file content. '
                                        '${base64:filename} to load binary file content and encode it with base64url. '
                                        '${env:environment_variable_name} to load the content of environment variable.')

automator_delete_parser = argparse.ArgumentParser(prog='automator-delete')
automator_delete_parser.add_argument('target', help='Automator ID or Name.')
automator_delete_parser.add_argument('--force', '-f', dest='force', action='store_true',
                                     help='Delete with no confirmation')

automator_action_parser = argparse.ArgumentParser(prog='automator-action')
automator_action_parser.add_argument('target', help='Automator ID or Name.')


def register_commands(commands):
    commands['automator'] = AutomatorCommand()


def register_command_info(_, command_info):
    command_info['automator'] = 'Manage Automator endpoints'


class AutomatorCommand(GroupCommand):
    def __init__(self):
        super(AutomatorCommand, self).__init__()
        self.register_command('list', AutomatorListCommand(), 'Displays a list of available automators.')
        self.register_command('view', AutomatorViewCommand(), 'Prints automator details.')
        self.register_command('create', AutomatorCreateCommand(), 'Creates automator endpoint.')
        self.register_command('edit', AutomatorEditCommand(), 'Changes automator configuration.')
        self.register_command('delete', AutomatorDeleteCommand(), 'Deletes automator.')
        self.register_command('setup', AutomatorSetupCommand(), 'Sets up automator.')
        action_command = AutomatorActionCommand()
        self.register_command('init', action_command, 'Initializes automator.')
        self.register_command('enable', action_command, 'Enables automator.')
        self.register_command('disable', action_command, 'Disables automator.')
        self.register_command('reset', action_command, 'Resets automator configuration to the default values.')
        action_command = AutomatorLogCommand()
        self.register_command('log', action_command, 'Retrieves automator logs.')
        self.register_command('log-clear', action_command, 'Clears automator logs.')

        self.default_verb = 'list'


class AutomatorMixin(object):
    @staticmethod
    def skill_to_name(skill):  # type: (automator_proto.SkillType) -> str
        if skill == automator_proto.DEVICE_APPROVAL:
            return 'Device Approval'
        elif skill == automator_proto.TEAM_APPROVAL:
            return 'Team Approval'
        else:
            return str(skill)

    @staticmethod
    def setting_to_str(setting):  # type: (automator_proto.AutomatorSettingValue) -> str
        return '{0}={1}'.format(setting.settingTag, setting.settingValue or '')

    @staticmethod
    def dump_automators(params):  # type: (KeeperParams) -> None
        table = []
        headers = ['ID', 'Name', 'Enabled', 'URL', 'Skills']
        if params.automators:
            for info in params.automators:
                row = [info.automatorId, info.name, info.enabled, info.url,
                       [AutomatorMixin.skill_to_name(x.skillType) for x in info.automatorSkills]]
                table.append(row)
        dump_report_data(table, headers=headers)

    @staticmethod
    def dump_automator(endpoint, status=False):    # type: (automator_proto.AutomatorInfo, bool) -> None
        logging.info('{0:>32s}: {1}'.format('Automator ID', endpoint.automatorId))
        logging.info('{0:>32s}: {1}'.format('Name', endpoint.name))
        logging.info('{0:>32s}: {1}'.format('URL', endpoint.url))
        logging.info('{0:>32s}: {1}'.format('Enabled', 'Yes' if endpoint.enabled else 'No'))
        if status:
            logging.info('{0:>32s}: {1}'.format('Initialized', 'Yes' if endpoint.status.initialized else 'No'))
            if endpoint.status.numberOfDevicesApproved > 0:
                logging.info('{0:>32}: {1}'.format('Approved Devices', endpoint.status.numberOfDevicesApproved))
            if endpoint.status.numberOfDevicesDenied > 0:
                logging.info('{0:>32}: {1}'.format('Denied Devices', endpoint.status.numberOfDevicesDenied))
            if endpoint.status.numberOfErrors > 0:
                logging.info('{0:>32}: {1}'.format('Number of Errors', endpoint.status.numberOfErrors))

        i = 0
        for skill in endpoint.automatorSkills:
            logging.info('{0:>32}: {1}'.format(
                'Skills' if i == 0 else '',
                AutomatorMixin.skill_to_name(skill.skillType)))
            i += 1

        visible_settings = [x for x in endpoint.automatorSettingValues if x.userVisible and x.editable]
        if visible_settings:
            logging.info('\n{0:>32s}\n'.format('Automator Settings'))
            for setting in visible_settings:
                setting_value = AutomatorMixin.setting_to_str(setting)
                value_len = len(setting_value)
                if value_len > 90:
                    value_len -= 80
                    setting_value = f'{setting_value[:80]}... {value_len} more characters.'

                logging.info('{0:>32s}: {1}'.format(setting.settingName, setting_value))

        if endpoint.logEntries:
            logging.info('\n{0:>32s}\n'.format('Automator Log'))
            for log in endpoint.logEntries:
                logging.info('<{0}> {1} - {2}'.format(log.messageLevel, log.serverTime, log.message))

    @staticmethod
    def ensure_loaded(params, force=False):   # type: (KeeperParams, bool) -> None
        if params.automators is None or force:
            enterprise_id = 0
            for node in params.enterprise['nodes']:
                enterprise_id = node['node_id'] >> 32
                break

            rq = automator_proto.AdminGetAutomatorsForEnterpriseRequest()
            rq.enterpriseId = enterprise_id
            admin_rs = api.communicate_rest(
                params, rq, 'automator/automator_get_all_for_enterprise', rs_type=automator_proto.AdminResponse)
            if admin_rs.success:
                params.automators = []
                params.automators.extend(admin_rs.automatorInfo)
            else:
                logging.warning(admin_rs.message)

    @staticmethod
    def find_automator(params, target):   # type: (KeeperParams, any) -> automator_proto.AutomatorInfo
        if not target:
            raise Exception('Automator Name cannot be empty')
        AutomatorMixin.ensure_loaded(params, False)
        if params.automators:
            try:
                target_id = int(target)
                for info in params.automators:
                    if target_id == info.automatorId:
                        return info
            except ValueError:
                pass

        name = target.lower() if type(target) == str else str(target)
        automators = [x for x in params.automators if x.name.lower() == name]
        if len(automators) == 0:
            raise Exception('Automator with name \"{0}\" not found'.format(target))
        if len(automators) > 1:
            raise Exception('There are %d automators with name \"{0})\". Use automator ID.'.format(target))
        return automators[0]


class AutomatorListCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return automator_list_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        self.ensure_loaded(params, True)
        self.dump_automators(params)


class AutomatorViewCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return automator_view_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        automator = self.find_automator(params, kwargs.get('target'))

        rq = automator_proto.AdminGetAutomatorRequest()
        rq.automatorId = automator.automatorId
        rs = api.communicate_rest(params, rq, 'automator/automator_get', rs_type=automator_proto.AdminResponse)

        self.dump_automator(rs.automatorInfo[0], status=True)
        if not rs.success:
            logging.info('')
            logging.info(rs.message)


class AutomatorCreateCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return automator_create_parser

    def execute(self, params, **kwargs):
        # type: (KeeperParams, **any) -> any
        name = kwargs.get('name')
        if not name:
            logging.warning('\"--name\" option is required for \"create\" command')
            return
        node = kwargs.get('node')
        nodes = list(self.resolve_nodes(params, node))
        if len(nodes) == 0:
            logging.warning('Node name \'%s\' is not found', node)
            return
        if len(nodes) > 1:
            logging.warning('Node name \'%s\' is not unique. Use Node ID.', node)
            return
        matched_node = nodes[0]

        rq = automator_proto.AdminCreateAutomatorRequest()
        rq.nodeId = matched_node['node_id']
        rq.name = name

        rs = api.communicate_rest(params, rq, 'automator/automator_create', rs_type=automator_proto.AdminResponse)
        if rs.success:
            self.dump_automator(rs.automatorInfo[0])
            params.automators = None
        else:
            logging.warning(rs.message)


class AutomatorEditCommand(EnterpriseCommand, AutomatorMixin):
    parameter_pattern = re.compile(r'^\${([^:]+?):([^}]+?)}$')

    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return automator_edit_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        automator = self.find_automator(params, kwargs.get('target'))
        rq = automator_proto.AdminEditAutomatorRequest()
        rq.automatorId = automator.automatorId
        name = kwargs['name']
        if name:
            rq.name = name
        url = kwargs['url']
        if url:
            rq.url = url
        skills = kwargs.get('skill')
        if skills:
            for skill in skills:
                if skill == 'device':
                    rq.skillTypes.append(automator_proto.DEVICE_APPROVAL)
                elif skill == 'team':
                    rq.skillTypes.append(automator_proto.TEAM_APPROVAL)
                else:
                    logging.warning('Unsupported skill: \"%s\"', skill)
                    return
        settings = kwargs.get('setting')
        if settings:
            available_settings = {}
            for sv in automator.automatorSettingValues:
                available_settings[sv.settingName.lower()] = sv

            for setting in settings:
                pos = setting.find('=')
                if pos > 1:
                    key = setting[:pos].strip()
                    existing = available_settings.get(key.lower())
                    if not existing:
                        logging.warning('Invalid property name: \"%s\"', key)
                        return

                    setting_value = automator_proto.AutomatorSettingValue()
                    setting_value.settingId = existing.settingId
                    value = setting[pos+1:].strip()
                    if existing.dataType == ssocloud.INTEGER or existing.dataType == ssocloud.LONG:
                        try:
                            _ = int(value)
                        except ValueError:
                            logging.warning('Numeric property \"%s\" has incorrect value: \"%s\"', key, value)
                            return
                    elif existing.dataType == ssocloud.BOOLEAN:
                        if value.lower() in ['true', '1', 'on']:
                            value = 'true'
                        elif value.lower() in ['false', '0', 'off']:
                            value = 'false'
                        else:
                            logging.warning(
                                'Boolean property \"%s\" has incorrect value: \"%s\". Use \"true\" or \"false\"',
                                key, value)
                            return

                    match = AutomatorEditCommand.parameter_pattern.match(value.strip())
                    if match:
                        parts = match.groups()
                        if len(parts) == 2:
                            if parts[0].lower() == 'file':
                                filename = parts[1].strip()
                                if os.path.isfile(filename):
                                    with open(filename, 'r') as f:
                                        value = f.read()
                            if parts[0].lower() == 'base64':
                                filename = parts[1].strip()
                                if os.path.isfile(filename):
                                    with open(filename, 'rb') as f:
                                        data = f.read()
                                        value = utils.base64_url_encode(data)
                            elif parts[0].lower() == 'env':
                                if parts[1] in os.environ:
                                    value = os.environ[parts[1]]

                    setting_value.settingValue = value
                    rq.automatorSettingValues.append(setting_value)
                else:
                    logging.warning('Incorrect setting format \"%s\". Expected KEY=VALUE', setting)
                    return

        rs = api.communicate_rest(params, rq, 'automator/automator_edit', rs_type=automator_proto.AdminResponse)
        if rs.automatorInfo:
            self.dump_automator(rs.automatorInfo[0])
            params.automators = None
        if not rs.success:
            logging.warning(rs.message)


class AutomatorSetupCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return automator_action_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        automator = self.find_automator(params, kwargs.get('target'))
        rq = automator_proto.AdminSetupAutomatorRequest()
        rq.automatorId = automator.automatorId
        rq.automatorState = automator_proto.NEEDS_CRYPTO_STEP_1
        rs = api.communicate_rest(params, rq, 'automator/automator_setup',
                                  rs_type=automator_proto.AdminSetupAutomatorResponse)
        if not rs.success:
            logging.warning('Automator \"%s\" setup step #1 error: %s', automator.name, rs.message)
            return
        rq = automator_proto.AdminSetupAutomatorRequest()
        rq.automatorId = automator.automatorId
        rq.automatorState = automator_proto.NEEDS_CRYPTO_STEP_2
        automator_public_key = crypto.load_ec_public_key(rs.automatorEcPublicKey)
        keys = params.enterprise['keys']
        if 'ecc_encrypted_private_key' in keys:
            encrypted_ec_private_key = utils.base64_url_decode(keys['ecc_encrypted_private_key'])
            ec_private_key = crypto.decrypt_aes_v2(encrypted_ec_private_key, params.enterprise['unencrypted_tree_key'])
            encrypted_ec_private_key = crypto.encrypt_ec(ec_private_key, automator_public_key)
            rq.encryptedEcEnterprisePrivateKey = encrypted_ec_private_key

        if 'rsa_encrypted_private_key' in keys:
            encrypted_rsa_private_key = utils.base64_url_decode(keys['rsa_encrypted_private_key'])
            rsa_private_key = crypto.decrypt_aes_v2(encrypted_rsa_private_key, params.enterprise['unencrypted_tree_key'])
            encrypted_rsa_private_key = crypto.encrypt_ec(rsa_private_key, automator_public_key)
            rq.encryptedRsaEnterprisePrivateKey = encrypted_rsa_private_key

        rs = api.communicate_rest(params, rq, 'automator/automator_setup',
                                  rs_type=automator_proto.AdminSetupAutomatorResponse)
        if rs.success:
            logging.info('Automator \"%s\" is setup', automator.name)
        else:
            logging.warning('Automator \"%s\" setup step #2 error: %s', automator.name, rs.message)


class AutomatorDeleteCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return automator_delete_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        automator = self.find_automator(params, kwargs.get('target'))
        rq = automator_proto.AdminDeleteAutomatorRequest()
        rq.automatorId = automator.automatorId
        _ = api.communicate_rest(params, rq, 'automator/automator_delete', rs_type=automator_proto.AdminResponse)
        logging.info('Automator %s deleted', automator.automatorId)
        params.automators = None


class AutomatorActionCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):  # type: () -> argparse.ArgumentParser | None
        return automator_action_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        automator = self.find_automator(params, kwargs.get('target'))
        action = kwargs.get('action') or ''
        if action == 'reset':
            rq = automator_proto.AdminDeleteAutomatorRequest()
            endpoint = 'automator/automator_full_reset'
        elif action == 'enable' or action == 'disable':
            rq = automator_proto.AdminEnableAutomatorRequest()
            rq.enabled = action == 'enable'
            endpoint = 'automator/automator_enable'
        elif action == 'init':
            rq = automator_proto.AdminInitializeAutomatorRequest()
            endpoint = 'automator/automator_initialize'
        else:
            logging.warning('Unsupported automator action \"%s\"', action)
            return
        rq.automatorId = automator.automatorId
        rs = api.communicate_rest(params, rq, endpoint, rs_type=automator_proto.AdminResponse)
        if rs.automatorInfo:
            self.dump_automator(rs.automatorInfo[0])
            params.automators = None
        if not rs.success:
            logging.warning(rs.message)


class AutomatorLogCommand(EnterpriseCommand, AutomatorMixin):
    def get_parser(self):  # type: () -> Optional[argparse.ArgumentParser]
        return automator_action_parser

    def execute(self, params, **kwargs):  # type: (KeeperParams, **any) -> any
        automator = self.find_automator(params, kwargs.get('target'))
        action = kwargs.get('action') or ''
        if action == 'log':
            rq = automator_proto.AdminAutomatorLogRequest()
            endpoint = 'automator/automator_log_get'
        elif action == 'log-clear':
            rq = automator_proto.AdminAutomatorLogClearRequest()
            endpoint = 'automator/automator_log_clear'
        else:
            logging.warning('Unsupported automator action \"%s\"', action)
            return
        rq.automatorId = automator.automatorId
        rs = api.communicate_rest(params, rq, endpoint, rs_type=automator_proto.AdminResponse)
        if rs.automatorInfo:
            self.dump_automator(rs.automatorInfo[0])



import tempfile
import json
import os
import warnings
from unittest import TestCase, mock

import pytest

from data_config import read_config_file
from keepercommander.params import KeeperParams
from keepercommander import cli, api, vault
from keepercommander.subfolder import BaseFolderNode
from keepercommander.commands import recordv3

@pytest.mark.integration
class TestConnectedCommands(TestCase):
    params = None

    @classmethod
    def setUpClass(cls):
        cls.params = KeeperParams()
        read_config_file(cls.params, 'vault.json')
        api.login(cls.params)
        TestConnectedCommands.wipe_out_data()

    @classmethod
    def tearDownClass(cls):
        cli.do_command(cls.params, 'logout')

    @classmethod
    def wipe_out_data(cls):
        params = cls.params    # type: KeeperParams

        params.revision = 0
        api.sync_down(params)
        request = {
            'command': 'record_update',
            'delete_records': [key for key in params.record_cache.keys()]
        }
        api.communicate(params, request)

        for shared_folder_uid in params.shared_folder_cache:
            request = {
                'command': 'shared_folder_update',
                'operation': 'delete',
                'shared_folder_uid': shared_folder_uid
            }
            api.communicate(params, request)

        folder_uids = [x for x in params.root_folder.subfolders if params.subfolder_cache[x]['type'] == BaseFolderNode.UserFolderType]
        if folder_uids:
            request = {
                'command': 'pre_delete',
                'objects': [
                    {
                        'from_type': 'user_folder',
                        'object_uid': x,
                        'object_type': 'user_folder',
                        'delete_resolution': 'unlink'
                    } for x in folder_uids
                ]
            }
            rs = api.communicate(params, request)
            request = {
                'command': 'delete',
                'pre_delete_token': rs['pre_delete_response']['pre_delete_token']
            }
            api.communicate(params, request)

        request = {
            'command': 'purge_deleted_records'
        }
        api.communicate(params, request)
        params.revision = 0
        api.sync_down(params)

    def setUp(self):
        warnings.simplefilter('ignore', category=ImportWarning)
        # Windows doesn't play nice with temp files, so create tmpdir that may fail to delete
        # self.tmpdir = os.path.join(os.path.dirname(__file__), 'tmp')
        # os.makedirs(self.tmpdir, exist_ok=True)
        # self.addCleanup(shutil.rmtree, self.tmpdir, ignore_errors=True)

    def test_commands(self):
        params = TestConnectedCommands.params # type: KeeperParams
        with mock.patch('builtins.input', side_effect=KeyboardInterrupt()), mock.patch('builtins.print'):
            record_uid = cli.do_command(params,
                'record-add --title="Record 1" --record-type=legacy login=user@keepersecurity.com password=$GEN url=https://keepersecurity.com/ cmdr:plugin=noop')
            cli.do_command(params, 'sync-down')

            rec = vault.KeeperRecord.load(params, record_uid)
            self.assertIsInstance(rec, vault.PasswordRecord)
            self.assertEqual(rec.get_custom_value('cmdr:plugin'), 'noop')
            old_password = rec.password
            cli.do_command(params, 'rotate -- {0}'.format(rec.record_uid))
            cli.do_command(params, 'sync-down')
            rec = vault.KeeperRecord.load(params, record_uid)
            self.assertIsInstance(rec, vault.PasswordRecord)
            self.assertNotEqual(old_password, rec.password)

            cli.do_command(params, 'ls -l')
            cli.do_command(params, 'mkdir --user-folder "User Folder 1"')
            cli.do_command(params, 'mkdir --shared-folder --all "Shared Folder 1"')
            cli.do_command(params, 'sync-down')
            cli.do_command(params, 'cd "User Folder 1"')
            cli.do_command(params, 'mkdir --user-folder "User Folder 2"')
            cli.do_command(params, 'cd /')
            cli.do_command(params, 'ln "Record 1" "Shared Folder 1"')
            cli.do_command(params, 'mv "Record 1" "User Folder 1"')
            params.revision = 0
            cli.do_command(params, 'sync-down')
            self.assertEqual(len(params.record_cache), 1)
            self.assertEqual(len(params.shared_folder_cache), 1)

            cli.do_command(params, 'cd "Shared Folder 1"')
            cli.do_command(params, 'append-notes --notes="Additional info" "Record 1"')
            cli.do_command(params, 'sync-down')
            cli.do_command(params, 'cd "../User Folder 1"')
            cli.do_command(params, 'rmdir --force "User Folder 2"')
            cli.do_command(params, 'sync-down')

            cli.do_command(params, 'cd /')
            cli.do_command(params, 'search record')
            cli.do_command(params, 'search folder')

            with tempfile.NamedTemporaryFile(delete=False) as f:
                try:
                    f.write(b'data')
                    f.flush()
                    cli.do_command(params, 'cd "User Folder 1"')
                    cli.do_command(params, 'upload-attachment --file="{0}" "Record 1"'.format(f.name))
                    f.close()
                finally:
                    os.remove(f.name)
            cli.do_command(params, 'sync-down')

            json_text = ''
            with mock.patch('builtins.open', mock.mock_open()) as m_open, mock.patch('os.path.abspath', return_value='file/path'):
                def file_write(text):
                    nonlocal json_text
                    json_text += text

                m_open.return_value.write = file_write
                cli.do_command(params, 'export --format=json file')
            self.assertTrue(len(json_text) > 0)
            exported = json.loads(json_text)
            if len(params.record_cache) > 0:
                self.assertEqual(len(params.record_cache), len(exported['records']))
            if len(params.shared_folder_cache) > 0:
                self.assertEqual(len(params.shared_folder_cache), len(exported['shared_folders']))

            rec = vault.KeeperRecord.load(params, record_uid)
            self.assertIsInstance(rec, vault.PasswordRecord)
            self.assertIsNotNone(rec.attachments)
            self.assertEqual(len(rec.attachments), 1)
            cli.do_command(params, 'delete-attachment --name={0} -- {1}'.format(rec.attachments[0].id, record_uid))
            cli.do_command(params, 'sync-down')
            rec = api.get_record(params, record_uid)
            self.assertEqual(len(rec.attachments), 0)

            script_path = os.path.dirname(__file__)
            cwd = os.getcwd()
            if script_path.startswith(cwd):
                script_path = script_path[len(cwd):]
                if script_path.startswith(os.sep):
                    script_path = script_path[1:]
            file = 'keepass.kdbx'
            if script_path:
                file = os.path.join(script_path, file)
            if os.path.isfile(file):
                os.remove(file)

            with mock.patch('getpass.getpass', return_value='password'), mock.patch('keepercommander.commands.base.user_choice', return_value='y'):
                cli.do_command(params, 'export --format=keepass "{0}"'.format(file))

            TestConnectedCommands.wipe_out_data()
            cli.do_command(params, 'sync-down')

            with mock.patch('getpass.getpass', return_value='password'), mock.patch('builtins.input', return_value=''):
                cli.do_command(params, 'import --format=keepass "{0}"'.format(file))
            cli.do_command(params, 'sync-down')

            TestConnectedCommands.wipe_out_data()
            with mock.patch('builtins.open', mock.mock_open()) as m_open, mock.patch('os.path.isfile', return_value=True):
                def file_read():
                    nonlocal json_text
                    return json_text

                m_open.return_value.read = file_read
                cli.do_command(params, 'import --format=json file')

            self.assertEqual(len(params.record_cache), len(exported['records']))
            self.assertEqual(len(params.shared_folder_cache), len(exported['shared_folders']))

    def test_quoting(self):
        """Check if quoting is self-consistent among mkdir and cd."""
        params = TestConnectedCommands.params
        with mock.patch('builtins.input', side_effect=KeyboardInterrupt()), mock.patch('builtins.print'):
            cli.do_command(params, 'mkdir --user-folder "/quoting"')
            cli.do_command(params, 'cd /quoting')
            for subdir in ('a//b', r'c\ d', r'e\"f', r"g\'h", r'\\'):
                cli.do_command(params, 'mkdir --user-folder {}'.format(subdir))
                cli.do_command(params, 'cd {}'.format(subdir))
                cli.do_command(params, 'cd /quoting')

    def test_vault_reports(self):
        params = TestConnectedCommands.params

        record_types = recordv3.RecordTypeInfo()
        report_json = record_types.execute(params, record_name='*', format='json')
        report = json.loads(report_json)
        self.assertTrue(isinstance(report, list))
        for record_type in report:
            type_name = record_type['content']
            report_json = record_types.execute(params, record_name=type_name, format='json')
            report = json.loads(report_json)
            self.assertTrue(isinstance(report, list))



import logging
import json
from datetime import datetime, timedelta
from unittest import TestCase, mock

from data_enterprise import EnterpriseEnvironment, get_enterprise_data, enterprise_allocate_ids
from keepercommander import api, crypto, utils
from keepercommander.record import Record
from keepercommander.params import KeeperParams
from keepercommander.error import CommandError
from data_vault import VaultEnvironment, get_connected_params
from keepercommander.commands import enterprise, aram


vault_env = VaultEnvironment()
ent_env = EnterpriseEnvironment()


class TestEnterprise(TestCase):
    expected_commands = []

    def setUp(self):
        TestEnterprise.use_data_key = True
        TestEnterprise.expected_commands.clear()
        self.communicate_mock = mock.patch('keepercommander.api.communicate').start()
        self.communicate_mock.side_effect = TestEnterprise.communicate_success
        self.query_enterprise_mock = mock.patch('keepercommander.api.query_enterprise').start()
        self.query_enterprise_mock.side_effect = TestEnterprise.query_enterprise

    def tearDown(self):
        mock.patch.stopall()

    def test_get_enterprise(self):
        params = get_connected_params()
        api.query_enterprise(params)
        self.assertIsNotNone(params.enterprise)
        self.assertEqual(params.enterprise['unencrypted_tree_key'], ent_env.tree_key)
        self.assertEqual(len(params.enterprise['nodes']), 2)

    def test_get_enterprise_public_key(self):
        TestEnterprise.use_data_key = False
        params = get_connected_params()
        api.query_enterprise(params)
        self.assertIsNotNone(params.enterprise)
        self.assertEqual(params.enterprise['unencrypted_tree_key'], ent_env.tree_key)
        self.assertEqual(len(params.enterprise['nodes']), 2)

    def test_enterprise_info_command(self):
        params = get_connected_params()
        api.query_enterprise(params)

        with mock.patch('builtins.print'):
            cmd = enterprise.EnterpriseInfoCommand()
            cmd.execute(params, verbose=True)

    def test_enterprise_add_user(self):
        params = get_connected_params()
        api.query_enterprise(params)

        cmd = enterprise.EnterpriseUserCommand()
        TestEnterprise.expected_commands = ['enterprise_user_add']
        cmd.execute(params, add=True, email='user2@keepercommander.com')
        self.assertEqual(len(TestEnterprise.expected_commands), 0)

    def test_enterprise_delete_user(self):
        params = get_connected_params()
        api.query_enterprise(params)

        cmd = enterprise.EnterpriseUserCommand()
        TestEnterprise.expected_commands = ['enterprise_user_delete']
        cmd.execute(params, delete=True, force=True, email=[ent_env.user2_email])
        self.assertEqual(len(TestEnterprise.expected_commands), 0)

    def test_enterprise_lock_user(self):
        params = get_connected_params()
        api.query_enterprise(params)

        cmd = enterprise.EnterpriseUserCommand()
        TestEnterprise.expected_commands = ['enterprise_user_lock', 'enterprise_user_lock']
        cmd.execute(params, unlock=True, email=[ent_env.user2_email])
        cmd.execute(params, lock=True, email=[ent_env.user2_email])
        self.assertEqual(len(TestEnterprise.expected_commands), 0)

    def test_enterprise_wrong_user(self):
        params = get_connected_params()
        api.query_enterprise(params)

        cmd = enterprise.EnterpriseUserCommand()
        with self.assertRaises(CommandError):
            with self.assertLogs(level=logging.WARNING):
                cmd.execute(params, lock=True, email=['wrong.user@keepersecurity.com'])

    def test_enterprise_expire_password(self):
        params = get_connected_params()
        api.query_enterprise(params)

        cmd = enterprise.EnterpriseUserCommand()
        TestEnterprise.expected_commands = ['set_master_password_expire']
        cmd.execute(params, expire=True, force=True, email=[ent_env.user2_email])
        self.assertEqual(len(TestEnterprise.expected_commands), 0)

        with mock.patch('keepercommander.commands.enterprise.user_choice') as mock_choice:
            TestEnterprise.expected_commands = ['set_master_password_expire']
            mock_choice.return_value = 'y'
            cmd.execute(params, expire=True, email=[ent_env.user2_email])
            with mock.patch('builtins.print'):
                self.assertEqual(len(TestEnterprise.expected_commands), 0)
                mock_choice.return_value = 'n'
                cmd.execute(params, expire=True, email=[ent_env.user2_email])

    def test_enterprise_user_update(self):
        params = get_connected_params()
        api.query_enterprise(params)

        cmd = enterprise.EnterpriseUserCommand()
        TestEnterprise.expected_commands = ['enterprise_user_update']
        cmd.execute(params, node='Enterprise 1', email=[ent_env.user2_email])
        self.assertEqual(len(TestEnterprise.expected_commands), 0)

        TestEnterprise.expected_commands = ['enterprise_user_update']
        cmd.execute(params, node='{0}'.format(ent_env.node1_id), email=[ent_env.user2_email])
        self.assertEqual(len(TestEnterprise.expected_commands), 0)

    def test_enterprise_user_team(self):
        params = get_connected_params()
        api.query_enterprise(params)

        cmd = enterprise.EnterpriseUserCommand()
        TestEnterprise.expected_commands = ['team_enterprise_user_add']
        cmd.execute(params, add_team=[ent_env.team1_uid], email=[ent_env.user2_email])
        self.assertEqual(len(TestEnterprise.expected_commands), 0)

        TestEnterprise.expected_commands = ['team_enterprise_user_remove']
        cmd.execute(params, remove_team=[ent_env.team1_uid], email=[ent_env.user2_email])
        self.assertEqual(len(TestEnterprise.expected_commands), 0)

    def test_enterprise_role(self):
        params = get_connected_params()
        api.query_enterprise(params)

        cmd = enterprise.EnterpriseRoleCommand()
        with mock.patch('builtins.print'):
            cmd.execute(params, role=[ent_env.role1_name])

        with mock.patch('keepercommander.commands.enterprise_common.user_choice') as mock_choice:
            mock_choice.return_value = 'y'
            TestEnterprise.expected_commands = ['role_user_add']
            cmd.execute(params, add_user=[ent_env.user2_email], role=[ent_env.role1_id])
            self.assertEqual(len(TestEnterprise.expected_commands), 0)

        TestEnterprise.expected_commands = ['role_user_remove']
        cmd.execute(params, remove_user=[ent_env.user2_email], role=[ent_env.role1_name])
        self.assertEqual(len(TestEnterprise.expected_commands), 0)

        with self.assertLogs(level=logging.WARNING):
            cmd.execute(params, add_user=[ent_env.user2_email], verbose=True, role=['Invalid'])
            with mock.patch('builtins.print'):
                cmd.execute(params, add_user=['invalid@keepersecurity.com'], verbose=True, role=[ent_env.role1_name])

    def test_enterprise_team(self):
        params = get_connected_params()
        api.query_enterprise(params)

        cmd = enterprise.EnterpriseTeamCommand()
        with mock.patch('builtins.print'):
            cmd.execute(params, team=[ent_env.team1_uid])

        TestEnterprise.expected_commands = ['team_add']
        cmd.execute(params, add=True, restrict_edit='on', node=str(ent_env.node1_id), team=['Team 3'])
        self.assertEqual(len(TestEnterprise.expected_commands), 0)

        with mock.patch('keepercommander.commands.enterprise.user_choice') as mock_choice:
            TestEnterprise.expected_commands = ['team_delete']
            mock_choice.return_value = 'y'
            cmd.execute(params, delete=True, team=['Team 1'])
            self.assertEqual(len(TestEnterprise.expected_commands), 0)

            with mock.patch('builtins.print'):
                mock_choice.return_value = 'n'
                cmd.execute(params, delete=True, team=[ent_env.team1_uid])
                self.assertEqual(len(TestEnterprise.expected_commands), 0)

        with self.assertLogs(level=logging.WARNING):
            cmd.execute(params, delete=True, team=['Unknown Team'])
            self.assertEqual(len(TestEnterprise.expected_commands), 0)

    def test_enterprise_team_user(self):
        params = get_connected_params()
        api.query_enterprise(params)

        cmd = enterprise.EnterpriseTeamCommand()

        TestEnterprise.expected_commands = ['team_enterprise_user_add']
        cmd.execute(params, add_user=[ent_env.user2_email], team=[ent_env.team1_uid])
        self.assertEqual(len(TestEnterprise.expected_commands), 0)

        TestEnterprise.expected_commands = ['team_enterprise_user_remove']
        cmd.execute(params, remove_user=[ent_env.user2_email], team=[ent_env.team1_uid])
        self.assertEqual(len(TestEnterprise.expected_commands), 0)

    def test_audit_log_splunk_properties_success(self):
        splunk = aram.AuditLogSplunkExport()
        props = {}
        record = Record()

        with mock.patch('builtins.print'), mock.patch('builtins.input') as mock_input, mock.patch('requests.post') as mock_post:
            resp1 = mock.Mock()
            resp1.status_code = 401
            resp1.json.return_value = {'code': 2}
            resp2 = mock.Mock()
            resp2.status_code = 400
            resp2.json.return_value = {'code': 6}
            mock_input.side_effect = ['www.splunk.com', 'Splunk Token', KeyboardInterrupt()]
            mock_post.side_effect = [resp1, resp2, Exception()]
            splunk.get_properties(record, props)
            self.assertIn('hec_url', props)
            self.assertIn('token', props)
            self.assertEqual(props['hec_url'], record.login_url)
            self.assertEqual(props['token'], record.password)
            self.assertTrue(splunk.store_record)

    def test_audit_log_splunk_properties_cancel(self):
        splunk = aram.AuditLogSplunkExport()
        props = {}
        record = Record()
        with mock.patch('builtins.print'), mock.patch('builtins.input') as mock_input, mock.patch('requests.post') as mock_post:
            resp1 = mock.Mock()
            resp1.status_code = 404
            mock_input.side_effect = ['www.splunk.com', KeyboardInterrupt()]
            mock_post.side_effect = [resp1, Exception()]
            with self.assertRaises(KeyboardInterrupt):
                splunk.get_properties(record, props)

    def test_audit_log_splunk_convert_event(self):
        splunk = aram.AuditLogSplunkExport()
        props = {
            'host': 'h',
            'enterprise_name': 'Unittest'
        }
        splunk.convert_event(props, self.get_audit_event())

    def test_audit_audit_report_parse_date_filter(self):
        cmd = aram.AuditReportCommand()

        epoch_max = int(datetime.now().timestamp())
        dt_max = datetime.fromtimestamp(epoch_max)

        dt_min = dt_max - timedelta(days=1)
        epoch_min = int(dt_min.timestamp())

        val = cmd.get_filter(dt_max.strftime('%Y-%m-%dT%H:%M:%SZ'), cmd.convert_date)
        self.assertTrue(type(val) == int)
        self.assertEqual(epoch_max, val)

        rng = cmd.get_filter('>{0}'.format(dt_min.strftime('%Y-%m-%dT%H:%M:%SZ')), cmd.convert_date)
        self.assertTrue(type(rng) == dict)
        self.assertIn('min', rng)
        self.assertIn('exclude_min', rng)
        self.assertTrue(rng['exclude_min'])
        self.assertNotIn('max', rng)
        self.assertEqual(rng['min'], epoch_min)

        rng = cmd.get_filter('<= {0}'.format(dt_max.strftime('%Y-%m-%dT%H:%M:%SZ')), cmd.convert_date)
        self.assertTrue(type(rng) == dict)
        self.assertIn('max', rng)
        self.assertFalse(rng.get('exclude_max') or False)
        self.assertNotIn('min', rng)
        self.assertEqual(rng['max'], epoch_max)

        rng = cmd.get_filter('between {0} and {1}'.format(dt_min.strftime('%Y-%m-%dT%H:%M:%SZ'), dt_max.strftime('%Y-%m-%dT%H:%M:%SZ')), cmd.convert_date)
        self.assertTrue(type(rng) == dict)
        self.assertIn('min', rng)
        self.assertIn('max', rng)
        self.assertEqual(rng['min'], epoch_min)
        self.assertEqual(rng['max'], epoch_max)

    def test_audit_audit_report_parse_int_filter(self):
        cmd = aram.AuditReportCommand()
        arr = cmd.get_filter('In (1,2,3, 4, 6,   5,7, 0)', cmd.convert_int)
        self.assertTrue(type(arr) == list)
        arr.sort()
        self.assertListEqual(arr, [0, 1, 2, 3, 4, 5, 6, 7])

    def test_enterprise_push_command(self):
        params = get_connected_params()
        api.query_enterprise(params)

        cmd = enterprise.EnterprisePushCommand()

        template_body = '''
[
    {
        "title": "Record For ${user_name}",
        "login": "${user_email}",
        "password": "${generate_password}",
        "login_url": "https://keepersecurity.com",
        "notes": "notes",
        "custom_fields": {
            "key1": "value1",
            "key2": "${user_email}"
        }
    },
    {
        "title": "Empty record"
    }

]'''
        templates = json.loads(template_body)
        values = {
            'user_name': api.generate_record_uid(),
            'generate_password': api.generate_record_uid(),
            'user_email': api.generate_record_uid()
        }
        cmd.enumerate_and_substitute_dict_fields(templates[0], values)
        cmd.enumerate_and_substitute_dict_fields(templates[1], values)
        self.assertEqual(templates[0]['title'], 'Record For {0}'.format(values['user_name']))
        self.assertEqual(templates[0]['password'], values['generate_password'])
        self.assertEqual(templates[0]['custom_fields']['key2'], values['user_email'])
        self.assertEqual(templates[1]['title'], 'Empty record')

        with self.assertRaises(CommandError):
            cmd.execute(params, file='template.json')

        with self.assertRaises(CommandError):
            cmd.execute(params, user=[ent_env.user2_email])

        def get_public_keys(_params, emails):
            for email in emails:
                emails[email] = vault_env.public_key

        with mock.patch('builtins.open', mock.mock_open(read_data=template_body)), \
                mock.patch('os.path.abspath', return_value='template.json'), \
                mock.patch('os.path.isfile', return_value=True), \
                mock.patch('keepercommander.commands.enterprise.EnterpriseCommand.get_public_keys') as m_pk:

            m_pk.side_effect = get_public_keys

            TestEnterprise.expected_commands = ['execute']
            cmd.execute(params, file='template.json', team=[ent_env.team2_name])
            self.assertEqual(len(TestEnterprise.expected_commands), 0)

            TestEnterprise.expected_commands = ['execute']
            cmd.execute(params, file='template.json', user=[ent_env.user2_email])
            self.assertEqual(len(TestEnterprise.expected_commands), 0)

    @staticmethod
    def get_audit_event():
        return {
            'id': 123456789098,
            'created': int(datetime.now().timestamp()),
            'username': vault_env.user,
            'ip_address': '9.9.9.9',
            'audit_event_type': 'login',
            'keeper_version': 'c14.0.0.0'
        }

    @staticmethod
    def query_enterprise(params):   # type: (KeeperParams) -> None
        params.enterprise = get_enterprise_data(params)
        if params.enterprise:
            encrypted_tree_key = utils.base64_url_decode(params.enterprise['tree_key'])
            params.enterprise['unencrypted_tree_key'] = crypto.decrypt_aes_v1(encrypted_tree_key, params.data_key)

            tree_key = params.enterprise['unencrypted_tree_key']
            for key in params.enterprise:
                o = params.enterprise[key]
                if not isinstance(o, list):
                    continue
                for elem in o:
                    if not isinstance(elem, dict):
                        continue
                    if 'encrypted_data' in elem:
                        decrypted_data = crypto.decrypt_aes_v1(utils.base64_url_decode(elem['encrypted_data']), tree_key)
                        elem['data'] = json.loads(decrypted_data.decode('utf-8'))

    @staticmethod
    def communicate_success(params, request):
        # type: (any, dict) -> dict
        if request['command'] == 'enterprise_allocate_ids':
            return enterprise_allocate_ids(params, request)

        rs = {
            'result': 'success',
            'result_code': '',
            'message': ''
        }
        if request['command'] == 'team_get_keys':
            rs['keys'] = [{
                'team_uid': x,
                'key': api.encrypt_aes(ent_env.team_key, vault_env.data_key),
                'type': 1
            } for x in request['teams']]
            return rs
        if request['command'] == 'public_keys':
            rs['public_keys'] = [{
                'key_owner': x,
                'public_key': vault_env.encoded_public_key
            } for x in request['key_owners']]
            return rs

        cmd = TestEnterprise.expected_commands.pop(0)
        if cmd == request['command']:
            return rs
        if request['command'] == 'execute':
            request = request['requests'][0]
            if cmd == request['command']:
                return rs
        raise Exception()



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper SDK
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import collections
from typing import Dict

FieldType = collections.namedtuple('FieldType', ['name', 'value', 'description'])

FieldTypes = {x[0]: x for x in (
    FieldType('text', '', 'plain text'),
    FieldType('url', '', 'url string, can be clicked'),
    FieldType('multiline', '', 'multiline text'),
    FieldType('fileRef', '', 'reference to the file field on another record'),
    FieldType('email', '', 'valid email address plus tag'),
    FieldType('host', {'hostName': '', 'port': ''}, 'multiple fields to capture host information'),
    FieldType('phone', {'region': '', 'number': '', 'ext': '', 'type': ''}, 'numbers and symbols only plus tag'),
    FieldType('name', {'first': '', 'middle': '', 'last': ''}, 'multiple fields to capture name'),
    FieldType('address', {'street1': '', 'street2': '', 'city': '', 'state': '', 'zip': '', 'country': ''},
              'multiple fields to capture address'),
    FieldType('addressRef', '', 'reference to the address field on another record'),
    FieldType('cardRef', '', 'reference to the card record type'),
    FieldType('secret', '', 'the field value is masked'),
    FieldType('login', '', 'Login field, detected as the website login for browser extension or KFFA.'),
    FieldType('password', '', 'Field value is masked and allows for generation. Also complexity enforcements.'),
    FieldType('securityQuestion', {'question': '', 'answer': ''}, 'Security Question and Answer'),
    FieldType('otp', '', 'captures the seed, displays QR code'),
    FieldType('paymentCard', {'cardNumber': '', 'cardExpirationDate': '', 'cardSecurityCode': ''},
              'Field consisting of validated card number, expiration date and security code.'),
    FieldType('date', 0, 'calendar date with validation, stored as unix milliseconds'),
    FieldType('bankAccount', {'accountType': '', 'routingNumber': '', 'accountNumber': ''},
              'bank account information'),
    FieldType('privateKey', {'publicKey': '', 'privateKey': ''}, 'private and/or public keys in ASN.1 format')
)}   # type: Dict[str, FieldType]


class Multiple:
    Never = 0
    Optional = 1
    Always = 2


RecordField = collections.namedtuple('RecordField', ['name', 'type', 'multiple'])
RecordFields = {x[0]: x for x in (
    RecordField('login', 'login', Multiple.Never),
    RecordField('password', 'password', Multiple.Never),
    RecordField('company', 'text', Multiple.Never),
    RecordField('licenseNumber', 'multiline', Multiple.Never),
    RecordField('accountNumber', 'text', Multiple.Never),
    RecordField('bankAccount', 'bankAccount', Multiple.Never),
    RecordField('note', 'multiline', Multiple.Never),
    RecordField('oneTimeCode', 'otp', Multiple.Never),
    RecordField('keyPair', 'privateKey', Multiple.Never),
    RecordField('pinCode', 'secret', Multiple.Never),
    RecordField('expirationDate', 'date', Multiple.Never),
    RecordField('birthDate', 'date', Multiple.Never),
    RecordField('name', 'name', Multiple.Optional),
    RecordField('phone', 'phone', Multiple.Optional),
    RecordField('email', 'email', Multiple.Optional),
    RecordField('address', 'address', Multiple.Optional),
    RecordField('date', 'date', Multiple.Optional),
    RecordField('paymentCard', 'paymentCard', Multiple.Optional),
    RecordField('multiline', 'multiline', Multiple.Optional),
    RecordField('url', 'url', Multiple.Optional),
    RecordField('host', 'host', Multiple.Optional),
    RecordField('secret', 'secret', Multiple.Optional),
    RecordField('otp', 'otp', Multiple.Optional),
    RecordField('text', 'text', Multiple.Optional),
    RecordField('securityQuestion', 'securityQuestion', Multiple.Always),
    RecordField('addressRef', 'addressRef', Multiple.Optional),
    RecordField('cardRef', 'cardRef', Multiple.Optional),
    RecordField('fileRef', 'fileRef', Multiple.Always),
)}   # type: Dict[str, RecordField]



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2021 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import json
import sys

from typing import List

from .. import imp_exp
from ..importer import (BaseFileImporter, BaseExporter, Record, RecordField, RecordSchemaField, RecordReferences, Folder, SharedFolder, Permission,
                        Team, BaseDownloadMembership)
from ... import api, utils, record_types
from ...proto import enterprise_pb2


class KeeperJsonImporter(BaseFileImporter):
    def do_import(self, filename, **kwargs):
        users_only = kwargs.get('users_only') or False
        with open(filename, "r", encoding='utf-8') as json_file:
            j = json.load(json_file)
            records = None
            folders = None
            if type(j) == list:
                records = j
            elif type(j) == dict:
                records = j.get('records')
                folders = j.get('shared_folders')
                teams = j.get('teams') if users_only else None

            if folders:
                for shf in folders:
                    fol = SharedFolder()
                    fol.uid = shf.get('uid')
                    fol.path = shf.get('path')
                    fol.manage_records = shf.get('manage_records') or False
                    fol.manage_users = shf.get('manage_users') or False
                    fol.can_edit = shf.get('can_edit') or False
                    fol.can_share = shf.get('can_share') or False
                    if users_only and 'permissions' in shf:
                        fol.permissions = []
                        permissions = shf['permissions']
                        if not isinstance(permissions, list):
                            permissions = [permissions]
                        for perm in permissions:
                            if isinstance(perm, dict):
                                p = Permission()
                                p.uid = perm.get('uid')
                                p.name = perm.get('name')
                                if p.uid or p.name:
                                    p.manage_records = perm.get('manage_records') or False
                                    p.manage_users = perm.get('manage_users') or False
                                    fol.permissions.append(p)

                    yield fol

            if isinstance(teams, list):
                for t in teams:
                    team = Team()
                    team.name = t.get('name')
                    if team.name:
                        team.uid = t.get('uid')
                        ms = t.get('members')
                        if isinstance(ms, list):
                            team.members = [x for x in ms if isinstance(x, str) and len(x) > 3]

                        yield team

            if not users_only and records:
                for r in records:
                    record = Record()
                    record.uid = r.get('uid')
                    if '$type' in r:
                        record.type = r['$type']
                    record.title = r.get('title')
                    record.login = r.get('login')
                    record.password = r.get('password')
                    record.login_url = r.get('login_url')
                    record.notes = r.get('notes')
                    custom_fields = r.get('custom_fields')
                    if type(custom_fields) is dict:
                        for name in custom_fields:
                            value = custom_fields[name]
                            if name[0] == '$':
                                pos = name.find(':')
                                if pos > 0:
                                    field_type = name[1:pos].strip()
                                    field_name = name[pos+1:].strip()
                                else:
                                    field_type = name[1:].strip()
                                    field_name = ''
                            else:
                                field_type = ''
                                field_name = name

                            ft = record_types.RecordFields.get(field_type or 'text')
                            if ft:
                                is_multiple = ft.multiple != record_types.Multiple.Never
                            else:
                                is_multiple = False
                                if not field_name:
                                    field_name = name
                                    field_type = ''

                            if isinstance(value, list) and not is_multiple:
                                for v in value:
                                    field = RecordField()
                                    field.type = field_type
                                    field.label = field_name
                                    field.value = v
                                    record.fields.append(field)
                            else:
                                field = RecordField()
                                field.type = field_type
                                field.label = field_name
                                field.value = value
                                record.fields.append(field)
                    if 'schema' in r:
                        record.schema = []
                        for s in r['schema']:
                            pos = s.find(':')
                            if pos > 0:
                                schema_ref = s[0:pos].strip()
                                schema_label = s[pos+1:].strip()
                            else:
                                schema_ref = s
                                schema_label = ''
                            if schema_ref[0] == '$':
                                schema_ref = schema_ref[1:]

                            sf = RecordSchemaField()
                            sf.ref = schema_ref
                            sf.label = schema_label
                            record.schema.append(sf)
                    if 'references' in r:
                        record.references = []
                        for ref_name in r['references']:
                            if not ref_name:
                                continue
                            ref_value = r['references'][ref_name]
                            if not ref_value:
                                continue
                            if type(ref_value) != list:
                                ref_value = [ref_value]
                            ref_type = ref_name
                            ref_label = ''
                            pos = field_name.find(':')
                            if pos > 0:
                                ref_type = field_name[1:pos].strip()
                                ref_label = field_name[pos+1].strip()
                            if ref_type[0] == '$':
                                ref_type = ref_type[1:]
                            rr = RecordReferences()
                            rr.type = ref_type
                            rr.label = ref_label
                            rr.uids.extend(ref_value)
                            record.references.append(rr)
                    if 'folders' in r:
                        record.folders = []
                        for f in r['folders']:
                            folder = Folder()
                            folder.domain = f.get('shared_folder')
                            folder.path = f.get('folder')
                            folder.can_edit = f.get('can_edit')
                            folder.can_share = f.get('can_share')
                            record.folders.append(folder)

                    yield record

    def extension(self):
        return 'json'


class KeeperJsonExporter(BaseExporter):

    def do_export(self, filename, items, file_password=None):
        shared_folders = []     # type: List[SharedFolder]
        records = []            # type: List[Record]
        teams = []              # type: List[Team]

        for item in items:
            if isinstance(item, Record):
                records.append(item)
            elif isinstance(item, SharedFolder):
                shared_folders.append(item)
            elif isinstance(item, Team):
                teams.append(item)
        """
        external_uids = {}
        external_id = 1
        for record in records:
            if record.uid:
                external_uids[record.uid] = external_id
                external_id += 1
        for record in records:
            if record.uid:
                record.uid = external_uids.get(record.uid)
            if record.references:
                for ref in record.references:
                    ref.uids = [external_uids[x] for x in ref.uids if x in external_uids]
        """
        ts = []
        for t in teams:
            team = {
                'name': t.name,
            }
            if t.uid:
                team['uid'] = t.uid
            if t.members:
                team['members'] = [x for x in t.members]
            ts.append(team)

        sfs = []
        for sf in shared_folders:
            sfo = {
                'path': sf.path,
            }
            if sf.uid:
                sfo['uid'] = sf.uid
            if sf.manage_users is not None:
                sfo['manage_users'] = sf.manage_users
            if sf.manage_records is not None:
                sfo['manage_records'] = sf.manage_records
            if sf.can_edit is not None:
                sfo['can_edit'] = sf.can_edit
            if sf.can_share is not None:
                sfo['can_share'] = sf.can_share

            if sf.permissions:
                sfo['permissions'] = []
                for perm in sf.permissions:
                    po = {
                        'name': perm.name,
                        'manage_users': perm.manage_users,
                        'manage_records': perm.manage_records
                    }
                    if perm.uid:
                        po['uid'] = perm.uid
                    sfo['permissions'].append(po)
            sfs.append(sfo)

        rs = []
        for r in records:
            ro = {
                'uid': r.uid or '',
                'title': r.title or ''
            }
            if r.login:
                ro['login'] = r.login
            if r.password:
                ro['password'] = r.password
            if r.login_url:
                ro['login_url'] = r.login_url
            if r.notes:
                ro['notes'] = r.notes
            if r.type:
                ro['$type'] = r.type
            if r.uid:
                ro['uid'] = r.uid

            if r.fields:
                ro['custom_fields'] = {}
                for field in r.fields:
                    if not field.type and field.label and field.label.startswith('$'):
                        field.type = 'text'
                    if field.type and field.label:
                        name = f'${field.type}:{field.label}'
                    elif field.type:
                        name = f'${field.type}'
                    else:
                        name = field.label or '<No Name>'
                    value = field.value
                    if name in ro['custom_fields']:
                        orig_value = ro['custom_fields'][name]
                        if orig_value:
                            orig_value = orig_value if type(orig_value) is list else [orig_value]
                        else:
                            orig_value = []
                        if value:
                            orig_value.append(value)
                        value = orig_value
                    ro['custom_fields'][name] = value

            if r.schema:
                ro['schema'] = []
                for rsf in r.schema:
                    name = f'${rsf.ref}'
                    if rsf.label:
                        name += f':{rsf.label}'
                    ro['schema'].append(name)

            if r.references:
                ro['references'] = {}
                for ref in r.references:
                    ref_name = f'${ref.type}:{ref.label}' if ref.type and ref.label else f'${ref.type}' if ref.type else ref.label or ''
                    refs = ro['references'].get(ref_name)
                    if refs is None:
                        refs = []
                        ro['references'][ref_name] = refs
                    refs.extend(ref.uids)

            if r.folders:
                ro['folders'] = []
                for folder in r.folders:
                    if folder.domain or folder.path:
                        fo = {}
                        ro['folders'].append(fo)
                        if folder.domain:
                            fo['shared_folder'] = folder.domain
                        if folder.path:
                            fo['folder'] = folder.path
                        if folder.can_edit:
                            fo['can_edit'] = True
                        if folder.can_share:
                            fo['can_share'] = True
            rs.append(ro)

        jo = {}
        if ts:
            jo['teams'] = ts
        if sfs:
            jo['shared_folders'] = sfs
        if rs:
            jo['records'] = rs

        if filename:
            with open(filename, mode="w", encoding='utf-8') as f:
                json.dump(jo, f, indent=2, ensure_ascii=False)
        else:
            json.dump(jo, sys.stdout, indent=2, ensure_ascii=False)
            print('')

    def has_shared_folders(self):
        return True

    def has_attachments(self):
        return False

    def extension(self):
        return 'json'

    def supports_stdout(self):
        return True

    def supports_v3_record(self):
        return True


class KeeperMembershipDownload(BaseDownloadMembership):
    def download_membership(self, params):
        teams = {}
        if params.shared_folder_cache:
            for shared_folder_uid in params.shared_folder_cache:
                shared_folder = api.get_shared_folder(params, shared_folder_uid)
                sf = SharedFolder()
                sf.uid = shared_folder.shared_folder_uid
                sf.path = imp_exp.get_folder_path(params, shared_folder.shared_folder_uid)
                sf.manage_users = shared_folder.default_manage_users
                sf.manage_records = shared_folder.default_manage_records
                sf.can_edit = shared_folder.default_can_edit
                sf.can_share = shared_folder.default_can_share
                sf.permissions = []
                if shared_folder.teams:
                    for team in shared_folder.teams:
                        perm = Permission()
                        perm.uid = team['team_uid']
                        perm.name = team['name']
                        perm.manage_users = team.get('manage_users', False)
                        perm.manage_records = team.get('manage_records', False)
                        teams[perm.uid] = perm.name
                        sf.permissions.append(perm)
                if shared_folder.users:
                    for user in shared_folder.users:
                        perm = Permission()
                        perm.name = user['username']
                        perm.manage_users = user.get('manage_users', False)
                        perm.manage_records = user.get('manage_records', False)
                        sf.permissions.append(perm)
                yield sf
        if teams and params.enterprise_ec_key:
            for team_uid in teams:
                t = Team()
                t.uid = team_uid
                t.name = teams[team_uid]
                rq = enterprise_pb2.GetTeamMemberRequest()
                rq.teamUid = utils.base64_url_decode(team_uid)
                rs = api.communicate_rest(params, rq, 'vault/get_team_members', rs_type=enterprise_pb2.GetTeamMemberResponse)
                t.members = [x.email for x in rs.enterpriseUser]
                yield t



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: BI.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x08\x42I.proto\x12\x02\x42I\"f\n\x1bValidateSessionTokenRequest\x12\x1d\n\x15\x65ncryptedSessionToken\x18\x01 \x01(\x0c\x12\x1c\n\x14returnMcEnterpiseIds\x18\x02 \x01(\x08\x12\n\n\x02ip\x18\x03 \x01(\t\"\xda\x02\n\x1cValidateSessionTokenResponse\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x0e\n\x06userId\x18\x02 \x01(\x05\x12\x18\n\x10\x65nterpriseUserId\x18\x03 \x01(\x03\x12\x37\n\x06status\x18\x04 \x01(\x0e\x32\'.BI.ValidateSessionTokenResponse.Status\x12\x15\n\rstatusMessage\x18\x05 \x01(\t\x12\x17\n\x0fmcEnterpriseIds\x18\x06 \x03(\x05\x12\x18\n\x10hasMSPPermission\x18\x07 \x01(\x08\x12\x1e\n\x16\x64\x65letedMcEnterpriseIds\x18\x08 \x03(\x05\"[\n\x06Status\x12\t\n\x05VALID\x10\x00\x12\r\n\tNOT_VALID\x10\x01\x12\x0b\n\x07\x45XPIRED\x10\x02\x12\x0e\n\nIP_BLOCKED\x10\x03\x12\x1a\n\x16INVALID_CLIENT_VERSION\x10\x04\"\x1b\n\x19SubscriptionStatusRequest\"\xf8\x01\n\x1aSubscriptionStatusResponse\x12$\n\x0b\x61utoRenewal\x18\x01 \x01(\x0b\x32\x0f.BI.AutoRenewal\x12/\n\x14\x63urrentPaymentMethod\x18\x02 \x01(\x0b\x32\x11.BI.PaymentMethod\x12\x14\n\x0c\x63heckoutLink\x18\x03 \x01(\t\x12\x19\n\x11licenseCreateDate\x18\x04 \x01(\x03\x12\x15\n\risDistributor\x18\x05 \x01(\x08\x12\x13\n\x0bisLegacyMsp\x18\x06 \x01(\x08\x12&\n\x0clicenseStats\x18\x08 \x03(\x0b\x32\x10.BI.LicenseStats\"\xd7\x01\n\x0cLicenseStats\x12#\n\x04type\x18\x01 \x01(\x0e\x32\x15.BI.LicenseStats.Type\x12\x11\n\tavailable\x18\x02 \x01(\x05\x12\x0c\n\x04used\x18\x03 \x01(\x05\"\x80\x01\n\x04Type\x12\x18\n\x14LICENSE_STAT_UNKNOWN\x10\x00\x12\x0c\n\x08MSP_BASE\x10\x01\x12\x0f\n\x0bMC_BUSINESS\x10\x02\x12\x14\n\x10MC_BUSINESS_PLUS\x10\x03\x12\x11\n\rMC_ENTERPRISE\x10\x04\x12\x16\n\x12MC_ENTERPRISE_PLUS\x10\x05\"@\n\x0b\x41utoRenewal\x12\x0e\n\x06nextOn\x18\x01 \x01(\x03\x12\x10\n\x08\x64\x61ysLeft\x18\x02 \x01(\x05\x12\x0f\n\x07isTrial\x18\x03 \x01(\x08\"\xcc\x02\n\rPaymentMethod\x12$\n\x04type\x18\x01 \x01(\x0e\x32\x16.BI.PaymentMethod.Type\x12$\n\x04\x63\x61rd\x18\x02 \x01(\x0b\x32\x16.BI.PaymentMethod.Card\x12$\n\x04sepa\x18\x03 \x01(\x0b\x32\x16.BI.PaymentMethod.Sepa\x12(\n\x06paypal\x18\x04 \x01(\x0b\x32\x18.BI.PaymentMethod.Paypal\x12\x15\n\rfailedBilling\x18\x05 \x01(\x08\x1a$\n\x04\x43\x61rd\x12\r\n\x05last4\x18\x01 \x01(\t\x12\r\n\x05\x62rand\x18\x02 \x01(\t\x1a&\n\x04Sepa\x12\r\n\x05last4\x18\x01 \x01(\t\x12\x0f\n\x07\x63ountry\x18\x02 \x01(\t\x1a\x08\n\x06Paypal\"0\n\x04Type\x12\x08\n\x04\x43\x41RD\x10\x00\x12\x08\n\x04SEPA\x10\x01\x12\n\n\x06PAYPAL\x10\x02\x12\x08\n\x04NONE\x10\x03\"\x1f\n\x1dSubscriptionMspPricingRequest\"\\\n\x1eSubscriptionMspPricingResponse\x12\x19\n\x06\x61\x64\x64ons\x18\x02 \x03(\x0b\x32\t.BI.Addon\x12\x1f\n\tfilePlans\x18\x03 \x03(\x0b\x32\x0c.BI.FilePlan\"\x1e\n\x1cSubscriptionMcPricingRequest\"|\n\x1dSubscriptionMcPricingResponse\x12\x1f\n\tbasePlans\x18\x01 \x03(\x0b\x32\x0c.BI.BasePlan\x12\x19\n\x06\x61\x64\x64ons\x18\x02 \x03(\x0b\x32\t.BI.Addon\x12\x1f\n\tfilePlans\x18\x03 \x03(\x0b\x32\x0c.BI.FilePlan\".\n\x08\x42\x61sePlan\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x16\n\x04\x63ost\x18\x02 \x01(\x0b\x32\x08.BI.Cost\"C\n\x05\x41\x64\x64on\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x16\n\x04\x63ost\x18\x02 \x01(\x0b\x32\x08.BI.Cost\x12\x16\n\x0e\x61mountConsumed\x18\x03 \x01(\x03\".\n\x08\x46ilePlan\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x16\n\x04\x63ost\x18\x02 \x01(\x0b\x32\x08.BI.Cost\"\xab\x01\n\x04\x43ost\x12\x0e\n\x06\x61mount\x18\x03 \x01(\x01\x12%\n\tamountPer\x18\x04 \x01(\x0e\x32\x12.BI.Cost.AmountPer\x12\x1e\n\x08\x63urrency\x18\x05 \x01(\x0e\x32\x0c.BI.Currency\"L\n\tAmountPer\x12\x0b\n\x07UNKNOWN\x10\x00\x12\t\n\x05MONTH\x10\x01\x12\x0e\n\nUSER_MONTH\x10\x02\x12\x17\n\x13USER_CONSUMED_MONTH\x10\x03\"=\n\x14InvoiceSearchRequest\x12\x0c\n\x04size\x18\x01 \x01(\x05\x12\x17\n\x0fstartingAfterId\x18\x02 \x01(\x05\"6\n\x15InvoiceSearchResponse\x12\x1d\n\x08invoices\x18\x01 \x03(\x0b\x32\x0b.BI.Invoice\"\xbe\x02\n\x07Invoice\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x15\n\rinvoiceNumber\x18\x02 \x01(\t\x12\x13\n\x0binvoiceDate\x18\x03 \x01(\x03\x12\x14\n\x0clicenseCount\x18\x04 \x01(\x05\x12#\n\ttotalCost\x18\x05 \x01(\x0b\x32\x10.BI.Invoice.Cost\x12%\n\x0binvoiceType\x18\x06 \x01(\x0e\x32\x10.BI.Invoice.Type\x1a\x36\n\x04\x43ost\x12\x0e\n\x06\x61mount\x18\x01 \x01(\x01\x12\x1e\n\x08\x63urrency\x18\x02 \x01(\x0e\x32\x0c.BI.Currency\"a\n\x04Type\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x07\n\x03NEW\x10\x01\x12\x0b\n\x07RENEWAL\x10\x02\x12\x0b\n\x07UPGRADE\x10\x03\x12\x0b\n\x07RESTORE\x10\x04\x12\x0f\n\x0b\x41SSOCIATION\x10\x05\x12\x0b\n\x07OVERAGE\x10\x06\"/\n\x16InvoiceDownloadRequest\x12\x15\n\rinvoiceNumber\x18\x01 \x01(\t\"9\n\x17InvoiceDownloadResponse\x12\x0c\n\x04link\x18\x01 \x01(\t\x12\x10\n\x08\x66ileName\x18\x02 \x01(\t\"<\n\x1dReportingDailySnapshotRequest\x12\r\n\x05month\x18\x01 \x01(\x05\x12\x0c\n\x04year\x18\x02 \x01(\x05\"v\n\x1eReportingDailySnapshotResponse\x12#\n\x07records\x18\x01 \x03(\x0b\x32\x12.BI.SnapshotRecord\x12/\n\rmcEnterprises\x18\x02 \x03(\x0b\x32\x18.BI.SnapshotMcEnterprise\"\xd7\x01\n\x0eSnapshotRecord\x12\x0c\n\x04\x64\x61te\x18\x01 \x01(\x03\x12\x16\n\x0emcEnterpriseId\x18\x02 \x01(\x05\x12\x17\n\x0fmaxLicenseCount\x18\x04 \x01(\x05\x12\x19\n\x11maxFilePlanTypeId\x18\x05 \x01(\x05\x12\x15\n\rmaxBasePlanId\x18\x06 \x01(\x05\x12(\n\x06\x61\x64\x64ons\x18\x07 \x03(\x0b\x32\x18.BI.SnapshotRecord.Addon\x1a*\n\x05\x41\x64\x64on\x12\x12\n\nmaxAddonId\x18\x01 \x01(\x05\x12\r\n\x05units\x18\x02 \x01(\x03\"0\n\x14SnapshotMcEnterprise\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t\"\x16\n\x14MappingAddonsRequest\"\\\n\x15MappingAddonsResponse\x12\x1f\n\x06\x61\x64\x64ons\x18\x01 \x03(\x0b\x32\x0f.BI.MappingItem\x12\"\n\tfilePlans\x18\x02 \x03(\x0b\x32\x0f.BI.MappingItem\"\'\n\x0bMappingItem\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0c\n\x04name\x18\x02 \x01(\t*D\n\x08\x43urrency\x12\x0b\n\x07UNKNOWN\x10\x00\x12\x07\n\x03USD\x10\x01\x12\x07\n\x03GBP\x10\x02\x12\x07\n\x03JPY\x10\x03\x12\x07\n\x03\x45UR\x10\x04\x12\x07\n\x03\x41UD\x10\x05\x42\x1e\n\x18\x63om.keepersecurity.protoB\x02\x42Ib\x06proto3')

_CURRENCY = DESCRIPTOR.enum_types_by_name['Currency']
Currency = enum_type_wrapper.EnumTypeWrapper(_CURRENCY)
UNKNOWN = 0
USD = 1
GBP = 2
JPY = 3
EUR = 4
AUD = 5


_VALIDATESESSIONTOKENREQUEST = DESCRIPTOR.message_types_by_name['ValidateSessionTokenRequest']
_VALIDATESESSIONTOKENRESPONSE = DESCRIPTOR.message_types_by_name['ValidateSessionTokenResponse']
_SUBSCRIPTIONSTATUSREQUEST = DESCRIPTOR.message_types_by_name['SubscriptionStatusRequest']
_SUBSCRIPTIONSTATUSRESPONSE = DESCRIPTOR.message_types_by_name['SubscriptionStatusResponse']
_LICENSESTATS = DESCRIPTOR.message_types_by_name['LicenseStats']
_AUTORENEWAL = DESCRIPTOR.message_types_by_name['AutoRenewal']
_PAYMENTMETHOD = DESCRIPTOR.message_types_by_name['PaymentMethod']
_PAYMENTMETHOD_CARD = _PAYMENTMETHOD.nested_types_by_name['Card']
_PAYMENTMETHOD_SEPA = _PAYMENTMETHOD.nested_types_by_name['Sepa']
_PAYMENTMETHOD_PAYPAL = _PAYMENTMETHOD.nested_types_by_name['Paypal']
_SUBSCRIPTIONMSPPRICINGREQUEST = DESCRIPTOR.message_types_by_name['SubscriptionMspPricingRequest']
_SUBSCRIPTIONMSPPRICINGRESPONSE = DESCRIPTOR.message_types_by_name['SubscriptionMspPricingResponse']
_SUBSCRIPTIONMCPRICINGREQUEST = DESCRIPTOR.message_types_by_name['SubscriptionMcPricingRequest']
_SUBSCRIPTIONMCPRICINGRESPONSE = DESCRIPTOR.message_types_by_name['SubscriptionMcPricingResponse']
_BASEPLAN = DESCRIPTOR.message_types_by_name['BasePlan']
_ADDON = DESCRIPTOR.message_types_by_name['Addon']
_FILEPLAN = DESCRIPTOR.message_types_by_name['FilePlan']
_COST = DESCRIPTOR.message_types_by_name['Cost']
_INVOICESEARCHREQUEST = DESCRIPTOR.message_types_by_name['InvoiceSearchRequest']
_INVOICESEARCHRESPONSE = DESCRIPTOR.message_types_by_name['InvoiceSearchResponse']
_INVOICE = DESCRIPTOR.message_types_by_name['Invoice']
_INVOICE_COST = _INVOICE.nested_types_by_name['Cost']
_INVOICEDOWNLOADREQUEST = DESCRIPTOR.message_types_by_name['InvoiceDownloadRequest']
_INVOICEDOWNLOADRESPONSE = DESCRIPTOR.message_types_by_name['InvoiceDownloadResponse']
_REPORTINGDAILYSNAPSHOTREQUEST = DESCRIPTOR.message_types_by_name['ReportingDailySnapshotRequest']
_REPORTINGDAILYSNAPSHOTRESPONSE = DESCRIPTOR.message_types_by_name['ReportingDailySnapshotResponse']
_SNAPSHOTRECORD = DESCRIPTOR.message_types_by_name['SnapshotRecord']
_SNAPSHOTRECORD_ADDON = _SNAPSHOTRECORD.nested_types_by_name['Addon']
_SNAPSHOTMCENTERPRISE = DESCRIPTOR.message_types_by_name['SnapshotMcEnterprise']
_MAPPINGADDONSREQUEST = DESCRIPTOR.message_types_by_name['MappingAddonsRequest']
_MAPPINGADDONSRESPONSE = DESCRIPTOR.message_types_by_name['MappingAddonsResponse']
_MAPPINGITEM = DESCRIPTOR.message_types_by_name['MappingItem']
_VALIDATESESSIONTOKENRESPONSE_STATUS = _VALIDATESESSIONTOKENRESPONSE.enum_types_by_name['Status']
_LICENSESTATS_TYPE = _LICENSESTATS.enum_types_by_name['Type']
_PAYMENTMETHOD_TYPE = _PAYMENTMETHOD.enum_types_by_name['Type']
_COST_AMOUNTPER = _COST.enum_types_by_name['AmountPer']
_INVOICE_TYPE = _INVOICE.enum_types_by_name['Type']
ValidateSessionTokenRequest = _reflection.GeneratedProtocolMessageType('ValidateSessionTokenRequest', (_message.Message,), {
  'DESCRIPTOR' : _VALIDATESESSIONTOKENREQUEST,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.ValidateSessionTokenRequest)
  })
_sym_db.RegisterMessage(ValidateSessionTokenRequest)

ValidateSessionTokenResponse = _reflection.GeneratedProtocolMessageType('ValidateSessionTokenResponse', (_message.Message,), {
  'DESCRIPTOR' : _VALIDATESESSIONTOKENRESPONSE,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.ValidateSessionTokenResponse)
  })
_sym_db.RegisterMessage(ValidateSessionTokenResponse)

SubscriptionStatusRequest = _reflection.GeneratedProtocolMessageType('SubscriptionStatusRequest', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIPTIONSTATUSREQUEST,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.SubscriptionStatusRequest)
  })
_sym_db.RegisterMessage(SubscriptionStatusRequest)

SubscriptionStatusResponse = _reflection.GeneratedProtocolMessageType('SubscriptionStatusResponse', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIPTIONSTATUSRESPONSE,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.SubscriptionStatusResponse)
  })
_sym_db.RegisterMessage(SubscriptionStatusResponse)

LicenseStats = _reflection.GeneratedProtocolMessageType('LicenseStats', (_message.Message,), {
  'DESCRIPTOR' : _LICENSESTATS,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.LicenseStats)
  })
_sym_db.RegisterMessage(LicenseStats)

AutoRenewal = _reflection.GeneratedProtocolMessageType('AutoRenewal', (_message.Message,), {
  'DESCRIPTOR' : _AUTORENEWAL,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.AutoRenewal)
  })
_sym_db.RegisterMessage(AutoRenewal)

PaymentMethod = _reflection.GeneratedProtocolMessageType('PaymentMethod', (_message.Message,), {

  'Card' : _reflection.GeneratedProtocolMessageType('Card', (_message.Message,), {
    'DESCRIPTOR' : _PAYMENTMETHOD_CARD,
    '__module__' : 'BI_pb2'
    # @@protoc_insertion_point(class_scope:BI.PaymentMethod.Card)
    })
  ,

  'Sepa' : _reflection.GeneratedProtocolMessageType('Sepa', (_message.Message,), {
    'DESCRIPTOR' : _PAYMENTMETHOD_SEPA,
    '__module__' : 'BI_pb2'
    # @@protoc_insertion_point(class_scope:BI.PaymentMethod.Sepa)
    })
  ,

  'Paypal' : _reflection.GeneratedProtocolMessageType('Paypal', (_message.Message,), {
    'DESCRIPTOR' : _PAYMENTMETHOD_PAYPAL,
    '__module__' : 'BI_pb2'
    # @@protoc_insertion_point(class_scope:BI.PaymentMethod.Paypal)
    })
  ,
  'DESCRIPTOR' : _PAYMENTMETHOD,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.PaymentMethod)
  })
_sym_db.RegisterMessage(PaymentMethod)
_sym_db.RegisterMessage(PaymentMethod.Card)
_sym_db.RegisterMessage(PaymentMethod.Sepa)
_sym_db.RegisterMessage(PaymentMethod.Paypal)

SubscriptionMspPricingRequest = _reflection.GeneratedProtocolMessageType('SubscriptionMspPricingRequest', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIPTIONMSPPRICINGREQUEST,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.SubscriptionMspPricingRequest)
  })
_sym_db.RegisterMessage(SubscriptionMspPricingRequest)

SubscriptionMspPricingResponse = _reflection.GeneratedProtocolMessageType('SubscriptionMspPricingResponse', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIPTIONMSPPRICINGRESPONSE,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.SubscriptionMspPricingResponse)
  })
_sym_db.RegisterMessage(SubscriptionMspPricingResponse)

SubscriptionMcPricingRequest = _reflection.GeneratedProtocolMessageType('SubscriptionMcPricingRequest', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIPTIONMCPRICINGREQUEST,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.SubscriptionMcPricingRequest)
  })
_sym_db.RegisterMessage(SubscriptionMcPricingRequest)

SubscriptionMcPricingResponse = _reflection.GeneratedProtocolMessageType('SubscriptionMcPricingResponse', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIPTIONMCPRICINGRESPONSE,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.SubscriptionMcPricingResponse)
  })
_sym_db.RegisterMessage(SubscriptionMcPricingResponse)

BasePlan = _reflection.GeneratedProtocolMessageType('BasePlan', (_message.Message,), {
  'DESCRIPTOR' : _BASEPLAN,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.BasePlan)
  })
_sym_db.RegisterMessage(BasePlan)

Addon = _reflection.GeneratedProtocolMessageType('Addon', (_message.Message,), {
  'DESCRIPTOR' : _ADDON,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.Addon)
  })
_sym_db.RegisterMessage(Addon)

FilePlan = _reflection.GeneratedProtocolMessageType('FilePlan', (_message.Message,), {
  'DESCRIPTOR' : _FILEPLAN,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.FilePlan)
  })
_sym_db.RegisterMessage(FilePlan)

Cost = _reflection.GeneratedProtocolMessageType('Cost', (_message.Message,), {
  'DESCRIPTOR' : _COST,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.Cost)
  })
_sym_db.RegisterMessage(Cost)

InvoiceSearchRequest = _reflection.GeneratedProtocolMessageType('InvoiceSearchRequest', (_message.Message,), {
  'DESCRIPTOR' : _INVOICESEARCHREQUEST,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.InvoiceSearchRequest)
  })
_sym_db.RegisterMessage(InvoiceSearchRequest)

InvoiceSearchResponse = _reflection.GeneratedProtocolMessageType('InvoiceSearchResponse', (_message.Message,), {
  'DESCRIPTOR' : _INVOICESEARCHRESPONSE,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.InvoiceSearchResponse)
  })
_sym_db.RegisterMessage(InvoiceSearchResponse)

Invoice = _reflection.GeneratedProtocolMessageType('Invoice', (_message.Message,), {

  'Cost' : _reflection.GeneratedProtocolMessageType('Cost', (_message.Message,), {
    'DESCRIPTOR' : _INVOICE_COST,
    '__module__' : 'BI_pb2'
    # @@protoc_insertion_point(class_scope:BI.Invoice.Cost)
    })
  ,
  'DESCRIPTOR' : _INVOICE,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.Invoice)
  })
_sym_db.RegisterMessage(Invoice)
_sym_db.RegisterMessage(Invoice.Cost)

InvoiceDownloadRequest = _reflection.GeneratedProtocolMessageType('InvoiceDownloadRequest', (_message.Message,), {
  'DESCRIPTOR' : _INVOICEDOWNLOADREQUEST,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.InvoiceDownloadRequest)
  })
_sym_db.RegisterMessage(InvoiceDownloadRequest)

InvoiceDownloadResponse = _reflection.GeneratedProtocolMessageType('InvoiceDownloadResponse', (_message.Message,), {
  'DESCRIPTOR' : _INVOICEDOWNLOADRESPONSE,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.InvoiceDownloadResponse)
  })
_sym_db.RegisterMessage(InvoiceDownloadResponse)

ReportingDailySnapshotRequest = _reflection.GeneratedProtocolMessageType('ReportingDailySnapshotRequest', (_message.Message,), {
  'DESCRIPTOR' : _REPORTINGDAILYSNAPSHOTREQUEST,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.ReportingDailySnapshotRequest)
  })
_sym_db.RegisterMessage(ReportingDailySnapshotRequest)

ReportingDailySnapshotResponse = _reflection.GeneratedProtocolMessageType('ReportingDailySnapshotResponse', (_message.Message,), {
  'DESCRIPTOR' : _REPORTINGDAILYSNAPSHOTRESPONSE,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.ReportingDailySnapshotResponse)
  })
_sym_db.RegisterMessage(ReportingDailySnapshotResponse)

SnapshotRecord = _reflection.GeneratedProtocolMessageType('SnapshotRecord', (_message.Message,), {

  'Addon' : _reflection.GeneratedProtocolMessageType('Addon', (_message.Message,), {
    'DESCRIPTOR' : _SNAPSHOTRECORD_ADDON,
    '__module__' : 'BI_pb2'
    # @@protoc_insertion_point(class_scope:BI.SnapshotRecord.Addon)
    })
  ,
  'DESCRIPTOR' : _SNAPSHOTRECORD,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.SnapshotRecord)
  })
_sym_db.RegisterMessage(SnapshotRecord)
_sym_db.RegisterMessage(SnapshotRecord.Addon)

SnapshotMcEnterprise = _reflection.GeneratedProtocolMessageType('SnapshotMcEnterprise', (_message.Message,), {
  'DESCRIPTOR' : _SNAPSHOTMCENTERPRISE,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.SnapshotMcEnterprise)
  })
_sym_db.RegisterMessage(SnapshotMcEnterprise)

MappingAddonsRequest = _reflection.GeneratedProtocolMessageType('MappingAddonsRequest', (_message.Message,), {
  'DESCRIPTOR' : _MAPPINGADDONSREQUEST,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.MappingAddonsRequest)
  })
_sym_db.RegisterMessage(MappingAddonsRequest)

MappingAddonsResponse = _reflection.GeneratedProtocolMessageType('MappingAddonsResponse', (_message.Message,), {
  'DESCRIPTOR' : _MAPPINGADDONSRESPONSE,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.MappingAddonsResponse)
  })
_sym_db.RegisterMessage(MappingAddonsResponse)

MappingItem = _reflection.GeneratedProtocolMessageType('MappingItem', (_message.Message,), {
  'DESCRIPTOR' : _MAPPINGITEM,
  '__module__' : 'BI_pb2'
  # @@protoc_insertion_point(class_scope:BI.MappingItem)
  })
_sym_db.RegisterMessage(MappingItem)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030com.keepersecurity.protoB\002BI'
  _CURRENCY._serialized_start=3149
  _CURRENCY._serialized_end=3217
  _VALIDATESESSIONTOKENREQUEST._serialized_start=16
  _VALIDATESESSIONTOKENREQUEST._serialized_end=118
  _VALIDATESESSIONTOKENRESPONSE._serialized_start=121
  _VALIDATESESSIONTOKENRESPONSE._serialized_end=467
  _VALIDATESESSIONTOKENRESPONSE_STATUS._serialized_start=376
  _VALIDATESESSIONTOKENRESPONSE_STATUS._serialized_end=467
  _SUBSCRIPTIONSTATUSREQUEST._serialized_start=469
  _SUBSCRIPTIONSTATUSREQUEST._serialized_end=496
  _SUBSCRIPTIONSTATUSRESPONSE._serialized_start=499
  _SUBSCRIPTIONSTATUSRESPONSE._serialized_end=747
  _LICENSESTATS._serialized_start=750
  _LICENSESTATS._serialized_end=965
  _LICENSESTATS_TYPE._serialized_start=837
  _LICENSESTATS_TYPE._serialized_end=965
  _AUTORENEWAL._serialized_start=967
  _AUTORENEWAL._serialized_end=1031
  _PAYMENTMETHOD._serialized_start=1034
  _PAYMENTMETHOD._serialized_end=1366
  _PAYMENTMETHOD_CARD._serialized_start=1230
  _PAYMENTMETHOD_CARD._serialized_end=1266
  _PAYMENTMETHOD_SEPA._serialized_start=1268
  _PAYMENTMETHOD_SEPA._serialized_end=1306
  _PAYMENTMETHOD_PAYPAL._serialized_start=1308
  _PAYMENTMETHOD_PAYPAL._serialized_end=1316
  _PAYMENTMETHOD_TYPE._serialized_start=1318
  _PAYMENTMETHOD_TYPE._serialized_end=1366
  _SUBSCRIPTIONMSPPRICINGREQUEST._serialized_start=1368
  _SUBSCRIPTIONMSPPRICINGREQUEST._serialized_end=1399
  _SUBSCRIPTIONMSPPRICINGRESPONSE._serialized_start=1401
  _SUBSCRIPTIONMSPPRICINGRESPONSE._serialized_end=1493
  _SUBSCRIPTIONMCPRICINGREQUEST._serialized_start=1495
  _SUBSCRIPTIONMCPRICINGREQUEST._serialized_end=1525
  _SUBSCRIPTIONMCPRICINGRESPONSE._serialized_start=1527
  _SUBSCRIPTIONMCPRICINGRESPONSE._serialized_end=1651
  _BASEPLAN._serialized_start=1653
  _BASEPLAN._serialized_end=1699
  _ADDON._serialized_start=1701
  _ADDON._serialized_end=1768
  _FILEPLAN._serialized_start=1770
  _FILEPLAN._serialized_end=1816
  _COST._serialized_start=1819
  _COST._serialized_end=1990
  _COST_AMOUNTPER._serialized_start=1914
  _COST_AMOUNTPER._serialized_end=1990
  _INVOICESEARCHREQUEST._serialized_start=1992
  _INVOICESEARCHREQUEST._serialized_end=2053
  _INVOICESEARCHRESPONSE._serialized_start=2055
  _INVOICESEARCHRESPONSE._serialized_end=2109
  _INVOICE._serialized_start=2112
  _INVOICE._serialized_end=2430
  _INVOICE_COST._serialized_start=2277
  _INVOICE_COST._serialized_end=2331
  _INVOICE_TYPE._serialized_start=2333
  _INVOICE_TYPE._serialized_end=2430
  _INVOICEDOWNLOADREQUEST._serialized_start=2432
  _INVOICEDOWNLOADREQUEST._serialized_end=2479
  _INVOICEDOWNLOADRESPONSE._serialized_start=2481
  _INVOICEDOWNLOADRESPONSE._serialized_end=2538
  _REPORTINGDAILYSNAPSHOTREQUEST._serialized_start=2540
  _REPORTINGDAILYSNAPSHOTREQUEST._serialized_end=2600
  _REPORTINGDAILYSNAPSHOTRESPONSE._serialized_start=2602
  _REPORTINGDAILYSNAPSHOTRESPONSE._serialized_end=2720
  _SNAPSHOTRECORD._serialized_start=2723
  _SNAPSHOTRECORD._serialized_end=2938
  _SNAPSHOTRECORD_ADDON._serialized_start=2896
  _SNAPSHOTRECORD_ADDON._serialized_end=2938
  _SNAPSHOTMCENTERPRISE._serialized_start=2940
  _SNAPSHOTMCENTERPRISE._serialized_end=2988
  _MAPPINGADDONSREQUEST._serialized_start=2990
  _MAPPINGADDONSREQUEST._serialized_end=3012
  _MAPPINGADDONSRESPONSE._serialized_start=3014
  _MAPPINGADDONSRESPONSE._serialized_end=3106
  _MAPPINGITEM._serialized_start=3108
  _MAPPINGITEM._serialized_end=3147
# @@protoc_insertion_point(module_scope)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#

import argparse
import fnmatch
import json
import logging
import re
from collections import OrderedDict
from typing import Optional, Tuple

from ..utils import is_url, is_email
from .base import raise_parse_exception, suppress_exit, Command
from .folder import get_folder_path
from .. import api, crypto, loginv3, utils
from ..params import KeeperParams
from ..proto import record_pb2
from ..subfolder import try_resolve_path, find_parent_top_folder


def register_commands(commands):
    commands['convert'] = ConvertCommand()


def register_command_info(aliases, command_info):
    command_info[convert_parser.prog] = convert_parser.description


convert_parser = argparse.ArgumentParser(prog='convert', description='Convert record(s) to use record types')
convert_parser.add_argument(
    '-t', '--record_type', dest='record_type', action='store', help='Convert to record type'
)
convert_parser.add_argument(
    '-q', '--quiet', dest='quiet', action='store_true', help="Don't display info about records matched and converted"
)
convert_parser.add_argument(
    '--ignore-ownership', dest='ignore_owner', action='store_true', help="Convert all records including not owned"
)
convert_parser.add_argument(
    '-f', '--force', dest='force', action='store_true',
    help='Convert all records including records with attachments. '
         'Please note that file attachments may not be decrypted by Keeper clients.'
)
convert_parser.add_argument(
    '-u', '--url', dest='url', action='store', help='Convert records with URL pattern (* for record with any URL)'
)
convert_parser.add_argument(
    '-n', '--dry-run', dest='dry_run', action='store_true', help='Preview the record conversions without updating'
)
convert_parser.add_argument(
    '-r', '--recursive', dest='recursive', action='store_true', help='Convert recursively through subfolders'
)
convert_parser.add_argument(
    'record-uid-name-patterns', nargs='*', type=str, action='store', help='One or more record title/UID search patterns'
)
convert_parser.error = raise_parse_exception
convert_parser.exit = suppress_exit


def get_matching_records_from_folder(params, folder_uid, regex, url_regex, attachments=False, ignore_owner=False):
    records = []
    if folder_uid in params.subfolder_record_cache:
        for uid in params.subfolder_record_cache[folder_uid]:
            if not ignore_owner:
                if uid not in params.meta_data_cache:
                    continue
                md = params.meta_data_cache[uid]
                if not md.get('owner', False):
                    continue
            if uid not in params.record_cache:
                continue
            rv = params.record_cache[uid].get('version', 0)
            if rv != 2:
                continue
            r = api.get_record(params, uid)
            if not attachments and r.attachments:
                continue
            r_attrs = (r.title, r.record_uid)
            if any(attr for attr in r_attrs if isinstance(attr, str) and len(attr) > 0 and regex(attr) is not None):
                if url_regex:
                    url_match = r.login_url and url_regex(r.login_url) is not None
                else:
                    url_match = True
                if url_match:
                    records.append(r)
    return records


def recurse_folder(params, folder_uid, folder_path, records_by_folder, regex, url_regex, recurse, attachments=False, ignore_owner=False):
    folder_records = get_matching_records_from_folder(params, folder_uid, regex, url_regex, attachments, ignore_owner)
    if len(folder_records) > 0:
        if folder_uid not in folder_path:
            folder_path[folder_uid] = get_folder_path(params, folder_uid)
            records_by_folder[folder_uid] = set()
        records_by_folder[folder_uid].update(folder_records)

    if recurse:
        folder = params.folder_cache[folder_uid] if folder_uid else params.root_folder
        for subfolder_uid in folder.subfolders:
            recurse_folder(params, subfolder_uid, folder_path, records_by_folder, regex, url_regex, recurse, attachments, ignore_owner)


class ConvertCommand(Command):
    def get_parser(self):
        return convert_parser

    def execute(self, params, **kwargs):
        if params.settings and isinstance(params.settings.get('record_types_enabled'), bool):
            v3_enabled = params.settings.get('record_types_enabled')
        else:
            v3_enabled = False
        if not v3_enabled:
            logging.warning(f'Cannot convert record(s) if record types is not enabled')
            return

        recurse = kwargs.get('recursive', False)
        url_pattern = kwargs.get('url')
        url_regex = re.compile(fnmatch.translate(url_pattern)).match if url_pattern else None

        record_type = kwargs.get('record_type') or 'login'
        available_types = [json.loads(x) for x in params.record_type_cache.values()]
        type_info = next((x for x in available_types if x.get('$id') == record_type), None)
        if type_info is None:
            logging.warning(
                f'Specified record type "{record_type}" is not valid. '
                f'Valid types are:\n{", ".join(sorted((x.get("$id") for x in available_types)))}'
            )
            return

        attachments = kwargs.get('force', False)
        if not isinstance(attachments, bool):
            attachments = False
        ignore_owner = kwargs.get('ignore_owner', False)
        if not isinstance(ignore_owner, bool):
            ignore_owner = False

        pattern_list = kwargs.get('record-uid-name-patterns', [])
        if len(pattern_list) == 0:
            logging.warning(f'Please specify a record to convert')
            return

        folder = params.folder_cache.get(params.current_folder, params.root_folder)
        records_by_folder = {}
        folder_path = {}
        for pattern in pattern_list:
            if pattern in params.record_cache:
                record = api.get_record(params, pattern)
                if record:
                    if '' not in folder_path:
                        folder_path[''] = ''
                    if '' not in records_by_folder:
                        records_by_folder[''] = set()
                    records_by_folder[''].add(record)
                continue
            rs = try_resolve_path(params, pattern)
            if rs is not None:
                folder, pattern = rs
            regex = re.compile(fnmatch.translate(pattern)).match if pattern else None

            folder_uid = folder.uid or ''
            recurse_folder(params, folder_uid, folder_path, records_by_folder, regex, url_regex, recurse,
                           attachments=attachments, ignore_owner=ignore_owner)

        if len(records_by_folder) == 0:
            patterns = ', '.join(pattern_list)
            msg = f'No records that can be converted to record types can be found for pattern "{patterns}"'
            if url_pattern:
                msg += f' with url matching "{url_pattern}"'
            logging.warning(msg)
            return

        # Sort records and if dry run print
        record_uids = set()
        record_names = OrderedDict()
        for folder_uid in sorted(folder_path, key=lambda x: folder_path[x]):
            path = folder_path[folder_uid]
            for record in sorted(records_by_folder[folder_uid], key=lambda x: getattr(x, 'title')):
                if record.record_uid not in record_uids:
                    record_uids.add(record.record_uid)
                    record_names[record.record_uid] = path + record.title

        dry_run = kwargs.get('dry_run', False)
        if dry_run:
            print(
                f'The following {len(record_uids)} record(s) that you own were matched'
                f' and would be converted to records with type "{record_type}":'
            )

            print('\n'.join(f' {k}  {v}' for k, v in record_names.items()))
        else:
            rq = {
                'command': 'sync_down',
                'revision': 0,
                'include': api.NON_SHARED_DATA_SCOPE + api.EXPLICIT
            }
            rs = api.communicate(params, rq)
            nsd = {x['record_uid']: x['data'] for x in rs.get('non_shared_data', [])}

            records = []
            for record_uid in record_uids:
                convert_result = ConvertCommand.convert_to_record_type_data(record_uid, params, type_info)
                if not convert_result:
                    logging.warning(f'Conversion failed for {record_names[record_uid]} ({record_uid})\n')
                    continue
                v3_data, file_info = convert_result
                record_key = params.record_cache[record_uid]['record_key_unencrypted']

                rc = record_pb2.RecordConvertToV3()
                rc.record_uid = loginv3.CommonHelperMethods.url_safe_str_to_bytes(record_uid)
                rc.client_modified_time = api.current_milli_time()
                record = api.get_record(params, record_uid)
                rc.revision = record.revision

                if file_info:
                    file_ref = next((x for x in v3_data['fields'] if x.get('type') == 'fileRef'), None)
                    if file_ref is None:
                        file_ref = {'type': 'fileRef'}
                        v3_data['fields'].append(file_ref)
                    if not isinstance(file_ref.get('value'), list):
                        file_ref['value'] = []

                    for f_info in file_info:
                        file_uid = utils.generate_uid()
                        file_ref['value'].append(file_uid)
                        file_key = utils.base64_url_decode(f_info['key'])

                        data = {}
                        for k in ('name', 'size', 'title', 'lastModified', 'type'):
                            data[k] = f_info[k]

                        rf = record_pb2.RecordFileForConversion()
                        rf.record_uid = utils.base64_url_decode(file_uid)
                        rf.file_file_id = f_info['id']
                        if 'thumbs' in f_info:
                            thumbs = f_info['thumbs']
                            if len(thumbs) > 0:
                                thumb = next((x for x in thumbs if isinstance(x, dict)), None)
                                if thumb:
                                    rf.thumb_file_id = thumbs[0]['id']
                        rf.data = crypto.encrypt_aes_v2(json.dumps(data).encode('utf-8'), file_key)
                        rf.record_key = crypto.encrypt_aes_v2(file_key, params.data_key)
                        rf.link_key = crypto.encrypt_aes_v2(file_key, record_key)
                        rc.record_file.append(rf)
                rc.data = crypto.encrypt_aes_v2(api.get_record_data_json_bytes(v3_data), record_key)

                # Non shared data
                if record_uid in nsd:
                    try:
                        non_shared_data = utils.base64_url_decode(nsd[record_uid])
                        non_shared_data = crypto.decrypt_aes_v1(non_shared_data, params.data_key)
                        rc.non_shared_data = crypto.encrypt_aes_v2(non_shared_data, params.data_key)
                    except Exception as e:
                        logging.debug('Non shared data conversion failed for record %s: ', record_uid, e)

                # Get share folder of the record so that we can convert the Record Folder Key
                shared_folders = find_parent_top_folder(params, record_uid)

                for shared_folder in shared_folders:
                    sf = params.shared_folder_cache[shared_folder.uid]
                    folder_key = record_pb2.RecordFolderForConversion()
                    folder_key.folder_uid = utils.base64_url_decode(shared_folder.uid)
                    folder_key.record_folder_key = crypto.encrypt_aes_v2(record_key, sf['shared_folder_key_unencrypted'])
                    rc.folder_key.append(folder_key)

                if params.enterprise_ec_key:
                    audit_data = {
                        'title': record.title or '',
                        'record_type': v3_data['type'],
                    }
                    if record.login_url:
                        audit_data['url'] = utils.url_strip(record.login_url)
                    rc.audit.data = crypto.encrypt_ec(json.dumps(audit_data).encode('utf-8'), params.enterprise_ec_key)

                records.append(rc)

            quiet = kwargs.get('quiet', False)
            if not quiet:
                logging.info(f'Matched {len(record_uids)} record(s)')

            if len(records) == 0:
                if not quiet:
                    logging.info('No records successfully converted')
                return

            while len(records) > 0:
                rq = record_pb2.RecordsConvertToV3Request()
                rq.records.extend(records[:999])
                records = records[999:]

                params.sync_data = True
                rq.client_time = api.current_milli_time()
                records_modify_rs = api.communicate_rest(params, rq, 'vault/records_convert3',
                                                         rs_type=record_pb2.RecordsModifyResponse)
                if not quiet:
                    converted_record_names = [
                        f' {utils.base64_url_encode(r.record_uid)}  {record_names[loginv3.CommonHelperMethods.bytes_to_url_safe_str(r.record_uid)]}'
                        for r in records_modify_rs.records if r.status == record_pb2.RS_SUCCESS
                    ]
                    if len(converted_record_names) > 0:
                        logging.info(f'Successfully converted the following {len(converted_record_names)} record(s):')
                        logging.info('\n'.join(converted_record_names))

                    convert_errors = [(f' {utils.base64_url_encode(x.record_uid)}  {record_names[loginv3.CommonHelperMethods.bytes_to_url_safe_str(x.record_uid)]}', x.message)
                                      for x in records_modify_rs.records if x.status != record_pb2.RS_SUCCESS]
                    if len(convert_errors) > 0:
                        logging.warning(f'Failed to convert the following {len(convert_errors)} record(s):')
                        logging.warning('\n'.join((f'{x[0]} : {x[1]}' for x in convert_errors)))

    @staticmethod
    def get_v3_field_type(field_value):
        return_type = 'text'
        if field_value:
            if is_url(field_value):
                return_type = 'url'
            elif is_email(field_value):
                return_type = 'email'
            if len(field_value) > 128:
                return_type = 'note'
        return return_type

    @staticmethod
    def convert_to_record_type_data(record_uid, params, type_info):
        # type: (str, KeeperParams, dict) -> Optional[Tuple[dict, list]]

        if not (record_uid and params and params.record_cache and record_uid in params.record_cache):
            logging.warning('Record %s not found.', record_uid)
            return

        record = params.record_cache[record_uid]
        version = record.get('version') or 0
        if version != 2:
            logging.warning('Record %s is not version 2.', record_uid)
            return

        data = record.get('data_unencrypted')
        extra = record.get('extra_unencrypted')

        data = data if isinstance(data, dict) else json.loads(data or '{}')
        extra = extra if isinstance(extra, dict) else json.loads(extra or '{}')

        # check for other non-convertible data - ex. fields[] has "field_type" != "totp" if present
        extra_fields = extra.get('fields') or []
        otps = [x['data'] for x in extra_fields if 'totp' == (x.get('field_type') or '') and 'data' in x]
        totp = otps[0] if len(otps) > 0 else ''
        otps = otps[1:]
        # label = otp.get('field_title') or ''

        title = data.get('title') or ''
        login = data.get('secret1') or ''
        password = data.get('secret2') or ''
        url = data.get('link') or ''
        v2_fields = {}
        if login:
            v2_fields['login'] = login
        if password:
            v2_fields['password'] = password
        if url:
            v2_fields['url'] = url
        if totp:
            v2_fields['oneTimeCode'] = totp

        notes = data.get('notes') or ''
        custom2 = data.get('custom') or []
        # custom.type	- Always "text" for legacy reasons.
        fields = []
        for field in type_info.get('fields', []):
            ref = field.get('$ref', 'text')
            label = field.get('label')
            typed_field = {
                'type': ref,
                'value': []
            }
            if label:
                typed_field['label'] = label
            if not label and ref in v2_fields:
                value = v2_fields.pop(ref)
                typed_field['value'].append(value)
            fields.append(typed_field)

        custom = []
        custom.extend(({
            'type': x[0],
            'value': [x[1]],
        } for x in v2_fields.items()))
        custom.extend(({
            'type': ConvertCommand.get_v3_field_type(x.get('value')),
            'label': x.get('name') or '',
            'value': [x.get('value')] if x.get('value') else []
        } for x in custom2 if x.get('name') or x.get('value')))

        # Add any remaining TOTP codes to custom[]
        if len(otps) > 0:
            custom.extend(({
                'type': 'oneTimeCode',
                'value': [x]
            } for x in otps if x))

        v3_data = {
            'title': title,
            'type': type_info['$id'],
            'fields': fields,
            'custom': custom,
            'notes': notes
        }

        file_info = extra.get('files') or []
        return v3_data, file_info



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: folder.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0c\x66older.proto\x12\x06\x46older\"\\\n\x10\x45ncryptedDataKey\x12\x14\n\x0c\x65ncryptedKey\x18\x01 \x01(\x0c\x12\x32\n\x10\x65ncryptedKeyType\x18\x02 \x01(\x0e\x32\x18.Folder.EncryptedKeyType\"\x82\x01\n\x16SharedFolderRecordData\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12\x0e\n\x06userId\x18\x03 \x01(\x05\x12\x32\n\x10\x65ncryptedDataKey\x18\x04 \x03(\x0b\x32\x18.Folder.EncryptedDataKey\"\\\n\x1aSharedFolderRecordDataList\x12>\n\x16sharedFolderRecordData\x18\x01 \x03(\x0b\x32\x1e.Folder.SharedFolderRecordData\"_\n\x15SharedFolderRecordFix\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x11\n\trecordUid\x18\x02 \x01(\x0c\x12 \n\x18\x65ncryptedRecordFolderKey\x18\x03 \x01(\x0c\"Y\n\x19SharedFolderRecordFixList\x12<\n\x15sharedFolderRecordFix\x18\x01 \x03(\x0b\x32\x1d.Folder.SharedFolderRecordFix\"\xa2\x02\n\rRecordRequest\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12&\n\nrecordType\x18\x02 \x01(\x0e\x32\x12.Folder.RecordType\x12\x12\n\nrecordData\x18\x03 \x01(\x0c\x12\x1a\n\x12\x65ncryptedRecordKey\x18\x04 \x01(\x0c\x12&\n\nfolderType\x18\x05 \x01(\x0e\x32\x12.Folder.FolderType\x12\x12\n\nhowLongAgo\x18\x06 \x01(\x03\x12\x11\n\tfolderUid\x18\x07 \x01(\x0c\x12 \n\x18\x65ncryptedRecordFolderKey\x18\x08 \x01(\x0c\x12\r\n\x05\x65xtra\x18\t \x01(\x0c\x12\x15\n\rnonSharedData\x18\n \x01(\x0c\x12\x0f\n\x07\x66ileIds\x18\x0b \x03(\x03\"E\n\x0eRecordResponse\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x0e\n\x06status\x18\x03 \x01(\t\"\x80\x01\n\x12SharedFolderFields\x12\x1b\n\x13\x65ncryptedFolderName\x18\x01 \x01(\x0c\x12\x13\n\x0bmanageUsers\x18\x02 \x01(\x08\x12\x15\n\rmanageRecords\x18\x03 \x01(\x08\x12\x0f\n\x07\x63\x61nEdit\x18\x04 \x01(\x08\x12\x10\n\x08\x63\x61nShare\x18\x05 \x01(\x08\"3\n\x18SharedFolderFolderFields\x12\x17\n\x0fsharedFolderUid\x18\x01 \x01(\x0c\"\x8f\x02\n\rFolderRequest\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12&\n\nfolderType\x18\x02 \x01(\x0e\x32\x12.Folder.FolderType\x12\x17\n\x0fparentFolderUid\x18\x03 \x01(\x0c\x12\x12\n\nfolderData\x18\x04 \x01(\x0c\x12\x1a\n\x12\x65ncryptedFolderKey\x18\x05 \x01(\x0c\x12\x36\n\x12sharedFolderFields\x18\x06 \x01(\x0b\x32\x1a.Folder.SharedFolderFields\x12\x42\n\x18sharedFolderFolderFields\x18\x07 \x01(\x0b\x32 .Folder.SharedFolderFolderFields\"E\n\x0e\x46olderResponse\x12\x11\n\tfolderUid\x18\x01 \x01(\x0c\x12\x10\n\x08revision\x18\x02 \x01(\x03\x12\x0e\n\x06status\x18\x03 \x01(\t\"w\n\x19ImportFolderRecordRequest\x12,\n\rfolderRequest\x18\x01 \x03(\x0b\x32\x15.Folder.FolderRequest\x12,\n\rrecordRequest\x18\x02 \x03(\x0b\x32\x15.Folder.RecordRequest\"|\n\x1aImportFolderRecordResponse\x12.\n\x0e\x66olderResponse\x18\x01 \x03(\x0b\x32\x16.Folder.FolderResponse\x12.\n\x0erecordResponse\x18\x02 \x03(\x0b\x32\x16.Folder.RecordResponse\"\xee\x01\n\x18SharedFolderUpdateRecord\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x17\n\x0fsharedFolderUid\x18\x02 \x01(\x0c\x12\x0f\n\x07teamUid\x18\x03 \x01(\x0c\x12(\n\x07\x63\x61nEdit\x18\x04 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12)\n\x08\x63\x61nShare\x18\x05 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12\x1a\n\x12\x65ncryptedRecordKey\x18\x06 \x01(\x0c\x12\x10\n\x08revision\x18\x07 \x01(\x05\x12\x12\n\nexpiration\x18\x08 \x01(\x12\"\xb5\x01\n\x16SharedFolderUpdateUser\x12\x10\n\x08username\x18\x01 \x01(\t\x12,\n\x0bmanageUsers\x18\x02 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12.\n\rmanageRecords\x18\x03 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12\x17\n\x0fsharedFolderKey\x18\x04 \x01(\x0c\x12\x12\n\nexpiration\x18\x05 \x01(\x12\"\x82\x01\n\x16SharedFolderUpdateTeam\x12\x0f\n\x07teamUid\x18\x01 \x01(\x0c\x12\x13\n\x0bmanageUsers\x18\x02 \x01(\x08\x12\x15\n\rmanageRecords\x18\x03 \x01(\x08\x12\x17\n\x0fsharedFolderKey\x18\x04 \x01(\x0c\x12\x12\n\nexpiration\x18\x05 \x01(\x12\"\x8e\x07\n\x1bSharedFolderUpdateV3Request\x12,\n$sharedFolderUpdateOperation_dont_use\x18\x01 \x01(\x05\x12\x17\n\x0fsharedFolderUid\x18\x02 \x01(\x0c\x12!\n\x19\x65ncryptedSharedFolderName\x18\x03 \x01(\x0c\x12\x10\n\x08revision\x18\x04 \x01(\x03\x12\x13\n\x0b\x66orceUpdate\x18\x05 \x01(\x08\x12\x13\n\x0b\x66romTeamUid\x18\x06 \x01(\x0c\x12\x33\n\x12\x64\x65\x66\x61ultManageUsers\x18\x07 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12\x35\n\x14\x64\x65\x66\x61ultManageRecords\x18\x08 \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12/\n\x0e\x64\x65\x66\x61ultCanEdit\x18\t \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12\x30\n\x0f\x64\x65\x66\x61ultCanShare\x18\n \x01(\x0e\x32\x17.Folder.SetBooleanValue\x12?\n\x15sharedFolderAddRecord\x18\x0b \x03(\x0b\x32 .Folder.SharedFolderUpdateRecord\x12;\n\x13sharedFolderAddUser\x18\x0c \x03(\x0b\x32\x1e.Folder.SharedFolderUpdateUser\x12;\n\x13sharedFolderAddTeam\x18\r \x03(\x0b\x32\x1e.Folder.SharedFolderUpdateTeam\x12\x42\n\x18sharedFolderUpdateRecord\x18\x0e \x03(\x0b\x32 .Folder.SharedFolderUpdateRecord\x12>\n\x16sharedFolderUpdateUser\x18\x0f \x03(\x0b\x32\x1e.Folder.SharedFolderUpdateUser\x12>\n\x16sharedFolderUpdateTeam\x18\x10 \x03(\x0b\x32\x1e.Folder.SharedFolderUpdateTeam\x12 \n\x18sharedFolderRemoveRecord\x18\x11 \x03(\x0c\x12\x1e\n\x16sharedFolderRemoveUser\x18\x12 \x03(\t\x12\x1e\n\x16sharedFolderRemoveTeam\x18\x13 \x03(\x0c\x12\x19\n\x11sharedFolderOwner\x18\x14 \x01(\t\"C\n\x1eSharedFolderUpdateRecordStatus\x12\x11\n\trecordUid\x18\x01 \x01(\x0c\x12\x0e\n\x06status\x18\x02 \x01(\t\"@\n\x1cSharedFolderUpdateUserStatus\x12\x10\n\x08username\x18\x01 \x01(\t\x12\x0e\n\x06status\x18\x02 \x01(\t\"?\n\x1cSharedFolderUpdateTeamStatus\x12\x0f\n\x07teamUid\x18\x01 \x01(\x0c\x12\x0e\n\x06status\x18\x02 \x01(\t\"\xdf\x05\n\x1cSharedFolderUpdateV3Response\x12\x10\n\x08revision\x18\x01 \x01(\x03\x12K\n\x1bsharedFolderAddRecordStatus\x18\x02 \x03(\x0b\x32&.Folder.SharedFolderUpdateRecordStatus\x12G\n\x19sharedFolderAddUserStatus\x18\x03 \x03(\x0b\x32$.Folder.SharedFolderUpdateUserStatus\x12G\n\x19sharedFolderAddTeamStatus\x18\x04 \x03(\x0b\x32$.Folder.SharedFolderUpdateTeamStatus\x12N\n\x1esharedFolderUpdateRecordStatus\x18\x05 \x03(\x0b\x32&.Folder.SharedFolderUpdateRecordStatus\x12J\n\x1csharedFolderUpdateUserStatus\x18\x06 \x03(\x0b\x32$.Folder.SharedFolderUpdateUserStatus\x12J\n\x1csharedFolderUpdateTeamStatus\x18\x07 \x03(\x0b\x32$.Folder.SharedFolderUpdateTeamStatus\x12N\n\x1esharedFolderRemoveRecordStatus\x18\x08 \x03(\x0b\x32&.Folder.SharedFolderUpdateRecordStatus\x12J\n\x1csharedFolderRemoveUserStatus\x18\t \x03(\x0b\x32$.Folder.SharedFolderUpdateUserStatus\x12J\n\x1csharedFolderRemoveTeamStatus\x18\n \x03(\x0b\x32$.Folder.SharedFolderUpdateTeamStatus*\x1a\n\nRecordType\x12\x0c\n\x08password\x10\x00*^\n\nFolderType\x12\x12\n\x0e\x64\x65\x66\x61ult_folder\x10\x00\x12\x0f\n\x0buser_folder\x10\x01\x12\x11\n\rshared_folder\x10\x02\x12\x18\n\x14shared_folder_folder\x10\x03*V\n\x10\x45ncryptedKeyType\x12\n\n\x06no_key\x10\x00\x12\x19\n\x15\x65ncrypted_by_data_key\x10\x01\x12\x1b\n\x17\x65ncrypted_by_public_key\x10\x02*M\n\x0fSetBooleanValue\x12\x15\n\x11\x42OOLEAN_NO_CHANGE\x10\x00\x12\x10\n\x0c\x42OOLEAN_TRUE\x10\x01\x12\x11\n\rBOOLEAN_FALSE\x10\x02\x42\"\n\x18\x63om.keepersecurity.protoB\x06\x46olderb\x06proto3')

_RECORDTYPE = DESCRIPTOR.enum_types_by_name['RecordType']
RecordType = enum_type_wrapper.EnumTypeWrapper(_RECORDTYPE)
_FOLDERTYPE = DESCRIPTOR.enum_types_by_name['FolderType']
FolderType = enum_type_wrapper.EnumTypeWrapper(_FOLDERTYPE)
_ENCRYPTEDKEYTYPE = DESCRIPTOR.enum_types_by_name['EncryptedKeyType']
EncryptedKeyType = enum_type_wrapper.EnumTypeWrapper(_ENCRYPTEDKEYTYPE)
_SETBOOLEANVALUE = DESCRIPTOR.enum_types_by_name['SetBooleanValue']
SetBooleanValue = enum_type_wrapper.EnumTypeWrapper(_SETBOOLEANVALUE)
password = 0
default_folder = 0
user_folder = 1
shared_folder = 2
shared_folder_folder = 3
no_key = 0
encrypted_by_data_key = 1
encrypted_by_public_key = 2
BOOLEAN_NO_CHANGE = 0
BOOLEAN_TRUE = 1
BOOLEAN_FALSE = 2


_ENCRYPTEDDATAKEY = DESCRIPTOR.message_types_by_name['EncryptedDataKey']
_SHAREDFOLDERRECORDDATA = DESCRIPTOR.message_types_by_name['SharedFolderRecordData']
_SHAREDFOLDERRECORDDATALIST = DESCRIPTOR.message_types_by_name['SharedFolderRecordDataList']
_SHAREDFOLDERRECORDFIX = DESCRIPTOR.message_types_by_name['SharedFolderRecordFix']
_SHAREDFOLDERRECORDFIXLIST = DESCRIPTOR.message_types_by_name['SharedFolderRecordFixList']
_RECORDREQUEST = DESCRIPTOR.message_types_by_name['RecordRequest']
_RECORDRESPONSE = DESCRIPTOR.message_types_by_name['RecordResponse']
_SHAREDFOLDERFIELDS = DESCRIPTOR.message_types_by_name['SharedFolderFields']
_SHAREDFOLDERFOLDERFIELDS = DESCRIPTOR.message_types_by_name['SharedFolderFolderFields']
_FOLDERREQUEST = DESCRIPTOR.message_types_by_name['FolderRequest']
_FOLDERRESPONSE = DESCRIPTOR.message_types_by_name['FolderResponse']
_IMPORTFOLDERRECORDREQUEST = DESCRIPTOR.message_types_by_name['ImportFolderRecordRequest']
_IMPORTFOLDERRECORDRESPONSE = DESCRIPTOR.message_types_by_name['ImportFolderRecordResponse']
_SHAREDFOLDERUPDATERECORD = DESCRIPTOR.message_types_by_name['SharedFolderUpdateRecord']
_SHAREDFOLDERUPDATEUSER = DESCRIPTOR.message_types_by_name['SharedFolderUpdateUser']
_SHAREDFOLDERUPDATETEAM = DESCRIPTOR.message_types_by_name['SharedFolderUpdateTeam']
_SHAREDFOLDERUPDATEV3REQUEST = DESCRIPTOR.message_types_by_name['SharedFolderUpdateV3Request']
_SHAREDFOLDERUPDATERECORDSTATUS = DESCRIPTOR.message_types_by_name['SharedFolderUpdateRecordStatus']
_SHAREDFOLDERUPDATEUSERSTATUS = DESCRIPTOR.message_types_by_name['SharedFolderUpdateUserStatus']
_SHAREDFOLDERUPDATETEAMSTATUS = DESCRIPTOR.message_types_by_name['SharedFolderUpdateTeamStatus']
_SHAREDFOLDERUPDATEV3RESPONSE = DESCRIPTOR.message_types_by_name['SharedFolderUpdateV3Response']
EncryptedDataKey = _reflection.GeneratedProtocolMessageType('EncryptedDataKey', (_message.Message,), {
  'DESCRIPTOR' : _ENCRYPTEDDATAKEY,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.EncryptedDataKey)
  })
_sym_db.RegisterMessage(EncryptedDataKey)

SharedFolderRecordData = _reflection.GeneratedProtocolMessageType('SharedFolderRecordData', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERRECORDDATA,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderRecordData)
  })
_sym_db.RegisterMessage(SharedFolderRecordData)

SharedFolderRecordDataList = _reflection.GeneratedProtocolMessageType('SharedFolderRecordDataList', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERRECORDDATALIST,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderRecordDataList)
  })
_sym_db.RegisterMessage(SharedFolderRecordDataList)

SharedFolderRecordFix = _reflection.GeneratedProtocolMessageType('SharedFolderRecordFix', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERRECORDFIX,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderRecordFix)
  })
_sym_db.RegisterMessage(SharedFolderRecordFix)

SharedFolderRecordFixList = _reflection.GeneratedProtocolMessageType('SharedFolderRecordFixList', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERRECORDFIXLIST,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderRecordFixList)
  })
_sym_db.RegisterMessage(SharedFolderRecordFixList)

RecordRequest = _reflection.GeneratedProtocolMessageType('RecordRequest', (_message.Message,), {
  'DESCRIPTOR' : _RECORDREQUEST,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.RecordRequest)
  })
_sym_db.RegisterMessage(RecordRequest)

RecordResponse = _reflection.GeneratedProtocolMessageType('RecordResponse', (_message.Message,), {
  'DESCRIPTOR' : _RECORDRESPONSE,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.RecordResponse)
  })
_sym_db.RegisterMessage(RecordResponse)

SharedFolderFields = _reflection.GeneratedProtocolMessageType('SharedFolderFields', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERFIELDS,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderFields)
  })
_sym_db.RegisterMessage(SharedFolderFields)

SharedFolderFolderFields = _reflection.GeneratedProtocolMessageType('SharedFolderFolderFields', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERFOLDERFIELDS,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderFolderFields)
  })
_sym_db.RegisterMessage(SharedFolderFolderFields)

FolderRequest = _reflection.GeneratedProtocolMessageType('FolderRequest', (_message.Message,), {
  'DESCRIPTOR' : _FOLDERREQUEST,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.FolderRequest)
  })
_sym_db.RegisterMessage(FolderRequest)

FolderResponse = _reflection.GeneratedProtocolMessageType('FolderResponse', (_message.Message,), {
  'DESCRIPTOR' : _FOLDERRESPONSE,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.FolderResponse)
  })
_sym_db.RegisterMessage(FolderResponse)

ImportFolderRecordRequest = _reflection.GeneratedProtocolMessageType('ImportFolderRecordRequest', (_message.Message,), {
  'DESCRIPTOR' : _IMPORTFOLDERRECORDREQUEST,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.ImportFolderRecordRequest)
  })
_sym_db.RegisterMessage(ImportFolderRecordRequest)

ImportFolderRecordResponse = _reflection.GeneratedProtocolMessageType('ImportFolderRecordResponse', (_message.Message,), {
  'DESCRIPTOR' : _IMPORTFOLDERRECORDRESPONSE,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.ImportFolderRecordResponse)
  })
_sym_db.RegisterMessage(ImportFolderRecordResponse)

SharedFolderUpdateRecord = _reflection.GeneratedProtocolMessageType('SharedFolderUpdateRecord', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERUPDATERECORD,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderUpdateRecord)
  })
_sym_db.RegisterMessage(SharedFolderUpdateRecord)

SharedFolderUpdateUser = _reflection.GeneratedProtocolMessageType('SharedFolderUpdateUser', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERUPDATEUSER,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderUpdateUser)
  })
_sym_db.RegisterMessage(SharedFolderUpdateUser)

SharedFolderUpdateTeam = _reflection.GeneratedProtocolMessageType('SharedFolderUpdateTeam', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERUPDATETEAM,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderUpdateTeam)
  })
_sym_db.RegisterMessage(SharedFolderUpdateTeam)

SharedFolderUpdateV3Request = _reflection.GeneratedProtocolMessageType('SharedFolderUpdateV3Request', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERUPDATEV3REQUEST,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderUpdateV3Request)
  })
_sym_db.RegisterMessage(SharedFolderUpdateV3Request)

SharedFolderUpdateRecordStatus = _reflection.GeneratedProtocolMessageType('SharedFolderUpdateRecordStatus', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERUPDATERECORDSTATUS,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderUpdateRecordStatus)
  })
_sym_db.RegisterMessage(SharedFolderUpdateRecordStatus)

SharedFolderUpdateUserStatus = _reflection.GeneratedProtocolMessageType('SharedFolderUpdateUserStatus', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERUPDATEUSERSTATUS,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderUpdateUserStatus)
  })
_sym_db.RegisterMessage(SharedFolderUpdateUserStatus)

SharedFolderUpdateTeamStatus = _reflection.GeneratedProtocolMessageType('SharedFolderUpdateTeamStatus', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERUPDATETEAMSTATUS,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderUpdateTeamStatus)
  })
_sym_db.RegisterMessage(SharedFolderUpdateTeamStatus)

SharedFolderUpdateV3Response = _reflection.GeneratedProtocolMessageType('SharedFolderUpdateV3Response', (_message.Message,), {
  'DESCRIPTOR' : _SHAREDFOLDERUPDATEV3RESPONSE,
  '__module__' : 'folder_pb2'
  # @@protoc_insertion_point(class_scope:Folder.SharedFolderUpdateV3Response)
  })
_sym_db.RegisterMessage(SharedFolderUpdateV3Response)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030com.keepersecurity.protoB\006Folder'
  _RECORDTYPE._serialized_start=4082
  _RECORDTYPE._serialized_end=4108
  _FOLDERTYPE._serialized_start=4110
  _FOLDERTYPE._serialized_end=4204
  _ENCRYPTEDKEYTYPE._serialized_start=4206
  _ENCRYPTEDKEYTYPE._serialized_end=4292
  _SETBOOLEANVALUE._serialized_start=4294
  _SETBOOLEANVALUE._serialized_end=4371
  _ENCRYPTEDDATAKEY._serialized_start=24
  _ENCRYPTEDDATAKEY._serialized_end=116
  _SHAREDFOLDERRECORDDATA._serialized_start=119
  _SHAREDFOLDERRECORDDATA._serialized_end=249
  _SHAREDFOLDERRECORDDATALIST._serialized_start=251
  _SHAREDFOLDERRECORDDATALIST._serialized_end=343
  _SHAREDFOLDERRECORDFIX._serialized_start=345
  _SHAREDFOLDERRECORDFIX._serialized_end=440
  _SHAREDFOLDERRECORDFIXLIST._serialized_start=442
  _SHAREDFOLDERRECORDFIXLIST._serialized_end=531
  _RECORDREQUEST._serialized_start=534
  _RECORDREQUEST._serialized_end=824
  _RECORDRESPONSE._serialized_start=826
  _RECORDRESPONSE._serialized_end=895
  _SHAREDFOLDERFIELDS._serialized_start=898
  _SHAREDFOLDERFIELDS._serialized_end=1026
  _SHAREDFOLDERFOLDERFIELDS._serialized_start=1028
  _SHAREDFOLDERFOLDERFIELDS._serialized_end=1079
  _FOLDERREQUEST._serialized_start=1082
  _FOLDERREQUEST._serialized_end=1353
  _FOLDERRESPONSE._serialized_start=1355
  _FOLDERRESPONSE._serialized_end=1424
  _IMPORTFOLDERRECORDREQUEST._serialized_start=1426
  _IMPORTFOLDERRECORDREQUEST._serialized_end=1545
  _IMPORTFOLDERRECORDRESPONSE._serialized_start=1547
  _IMPORTFOLDERRECORDRESPONSE._serialized_end=1671
  _SHAREDFOLDERUPDATERECORD._serialized_start=1674
  _SHAREDFOLDERUPDATERECORD._serialized_end=1912
  _SHAREDFOLDERUPDATEUSER._serialized_start=1915
  _SHAREDFOLDERUPDATEUSER._serialized_end=2096
  _SHAREDFOLDERUPDATETEAM._serialized_start=2099
  _SHAREDFOLDERUPDATETEAM._serialized_end=2229
  _SHAREDFOLDERUPDATEV3REQUEST._serialized_start=2232
  _SHAREDFOLDERUPDATEV3REQUEST._serialized_end=3142
  _SHAREDFOLDERUPDATERECORDSTATUS._serialized_start=3144
  _SHAREDFOLDERUPDATERECORDSTATUS._serialized_end=3211
  _SHAREDFOLDERUPDATEUSERSTATUS._serialized_start=3213
  _SHAREDFOLDERUPDATEUSERSTATUS._serialized_end=3277
  _SHAREDFOLDERUPDATETEAMSTATUS._serialized_start=3279
  _SHAREDFOLDERUPDATETEAMSTATUS._serialized_end=3342
  _SHAREDFOLDERUPDATEV3RESPONSE._serialized_start=3345
  _SHAREDFOLDERUPDATEV3RESPONSE._serialized_end=4080
# @@protoc_insertion_point(module_scope)



#  _  __
# | |/ /___ ___ _ __  ___ _ _ ®
# | ' </ -_) -_) '_ \/ -_) '_|
# |_|\_\___\___| .__/\___|_|
#              |_|
#
# Keeper Commander
# Copyright 2022 Keeper Security Inc.
# Contact: ops@keepersecurity.com
#
import requests
from contextlib import contextmanager

from ..importer import Attachment


class ManageEngineAttachment(Attachment):
    def __init__(self, name, request_kwargs):
        Attachment.__init__(self)
        self.name = name

        attach_size = 0
        request_kwargs['stream'] = True
        with requests.get(**request_kwargs) as r:
            r.raw.decode_content = True
            for chunk in r.iter_content(chunk_size=10000):
                attach_size += len(chunk)

        self.size = attach_size
        self.request_kwargs = request_kwargs

    @contextmanager
    def open(self):
        response = requests.get(**self.request_kwargs)
        response.raw.decode_content = True
        yield response.raw



from unittest import TestCase, mock

from data_vault import VaultEnvironment, get_connected_params, get_sync_down_response, get_synced_params
from helper import KeeperApiHelper
from keepercommander.api import sync_down

vault_env = VaultEnvironment()


class TestSyncDown(TestCase):

    def setUp(self):
        self.communicate_mock = mock.patch('keepercommander.api.communicate').start()
        self.communicate_mock.side_effect = KeeperApiHelper.communicate_command

    def tearDown(self):
        mock.patch.stopall()

    def test_full_sync(self):
        params = get_connected_params()
        self.communicate_mock.side_effect = None
        self.communicate_mock.return_value = get_sync_down_response()

        sync_down(params)

        self.assertEqual(len(params.record_cache), 3)
        self.assertEqual(len(params.shared_folder_cache), 1)
        self.assertEqual(len(params.team_cache), 1)
        self.assert_key_unencrypted(params)

    def test_sync_remove_owned_records(self):
        params = get_synced_params()
        len_before = len(params.record_cache)

        records_to_delete = [x['record_uid'] for x in params.meta_data_cache.values() if x['owner']]

        def sync_down_removed_records(rq):
            self.assertEqual(rq['command'], 'sync_down')
            return {
                'revision': vault_env.revision + 1,
                'removed_records': records_to_delete
            }

        KeeperApiHelper.communicate_expect([sync_down_removed_records])
        sync_down(params)
        self.assertTrue(KeeperApiHelper.is_expect_empty())

        self.assertEqual(len(params.record_cache), len_before - len(records_to_delete))
        self.assert_key_unencrypted(params)

    def test_sync_remove_team(self):
        params = get_synced_params()
        teams_to_delete = [x['team_uid'] for x in params.team_cache.values()]

        def sync_down_removed_teams(rq):
            self.assertEqual(rq['command'], 'sync_down')
            return {
                'revision': vault_env.revision + 1,
                'removed_teams': teams_to_delete
            }

        KeeperApiHelper.communicate_expect([sync_down_removed_teams])
        sync_down(params)
        self.assertTrue(KeeperApiHelper.is_expect_empty())

        self.assertEqual(len(params.record_cache), 3)
        self.assertEqual(len(params.team_cache), 0)
        self.assert_key_unencrypted(params)

    def test_sync_remove_shared_folder_then_team(self):
        params = get_synced_params()
        sf_to_delete = [x['shared_folder_uid'] for x in params.shared_folder_cache.values()]

        def sync_down_removed_shared_folders(rq):
            self.assertEqual(rq['command'], 'sync_down')
            return {
                'revision': vault_env.revision + 1,
                'removed_shared_folders': sf_to_delete
            }

        KeeperApiHelper.communicate_expect([sync_down_removed_shared_folders])
        sync_down(params)
        self.assertTrue(KeeperApiHelper.is_expect_empty())

        self.assertEqual(len(params.record_cache), 3)
        self.assertEqual(len(params.shared_folder_cache), 1)
        self.assertEqual(len(params.team_cache), 1)
        self.assert_key_unencrypted(params)

        teams_to_delete = [x['team_uid'] for x in params.team_cache.values()]

        def sync_down_removed_teams(rq):
            self.assertEqual(rq['command'], 'sync_down')
            return {
                'revision': vault_env.revision + 1,
                'removed_teams': teams_to_delete
            }

        KeeperApiHelper.communicate_expect([sync_down_removed_teams])
        sync_down(params)
        self.assertTrue(KeeperApiHelper.is_expect_empty())

        self.assertEqual(len(params.record_cache), 2)
        self.assertEqual(len(params.shared_folder_cache), 0)
        self.assertEqual(len(params.team_cache), 0)
        self.assert_key_unencrypted(params)

    def test_sync_remove_team_shared_folder(self):
        params = get_synced_params()
        teams_to_delete = [x['team_uid'] for x in params.team_cache.values()]
        sf_to_delete = [x['shared_folder_uid'] for x in params.shared_folder_cache.values()]

        def sync_down_removed_teams_and_shared_folders(rq):
            self.assertEqual(rq['command'], 'sync_down')
            return {
                'revision': vault_env.revision + 1,
                'removed_shared_folders': sf_to_delete,
                'removed_teams': teams_to_delete
            }

        KeeperApiHelper.communicate_expect([sync_down_removed_teams_and_shared_folders])
        sync_down(params)
        self.assertTrue(KeeperApiHelper.is_expect_empty())

        self.assertEqual(len(params.record_cache), 2)
        self.assertEqual(len(params.shared_folder_cache), 0)
        self.assertEqual(len(params.team_cache), 0)
        self.assert_key_unencrypted(params)

    def assert_key_unencrypted(self, params):
        for r in params.record_cache.values():
            self.assertTrue('record_key_unencrypted' in r)
        for sf in params.shared_folder_cache.values():
            self.assertTrue('shared_folder_key_unencrypted' in sf)
        for t in params.team_cache.values():
            self.assertTrue('team_key_unencrypted' in t)
